import argparse
import os
import sys
from .scripts.find_DMRs import main as fdmr_main
import time
import multiprocessing as mp

def call_fdmr(args):
    """Calls find_DMRs with one prepare_for_DMR_finding-file"""
    prepfile, test_method, fdr, cutoff, plot, folder_out = args
    start = time.time()
    out = fdmr_main(prepfile, test_method, fdr, cutoff, plot, folder_out)
    end = time.time()
    print(prepfile+' -- '+str(end-start)+' seconds')

    return out #out_filename, num_MRs, num_DMRs, num_hyper, num_hypo

def main(listfile, cutoff, test_method, fdr, plot, num_processes, folder_out):

    if test_method not in ('Rranksum', 'Pranksum', 'Mranksum', 'Ttest','KStest'):
        print('***** ERROR in find_DMRs.py: testMethod must be either Ttest, KStest, Mranksum, Rranksum or Pranksum. Exiting *****')
        exit(1)

    if 'imputed' in listfile:
        out_list_filename = folder_out+'/list_DMR_files_imputed_'+test_method+'.txt'
        out_DMR_counts_mC = folder_out+'/counts_DMR_hypo_hyper_imputed_'+test_method+'_5mC.csv'
        out_DMR_counts_hmC = folder_out+'/counts_DMR_hypo_hyper_imputed_'+test_method+'_5hmC.csv'
    elif 'increased' in listfile:
        out_list_filename = folder_out+'/list_DMR_files_increased_'+test_method+'.txt'
        out_DMR_counts_mC = folder_out+'/counts_DMR_hypo_hyper_increased_'+test_method+'_5mC.csv'
        out_DMR_counts_hmC = folder_out+'/counts_DMR_hypo_hyper_increased_'+test_method+'_5hmC.csv'
    else:
        print('***** ERROR in DMR_search.py: filename list_prepared_for_DMR_finding_... Filename must contain either word, imputed or increased. Exiting *****')
        exit(1)

    out_list_file = open(out_list_filename, 'w')
    out_count_mC = open(out_DMR_counts_mC, 'w')
    out_count_hmC = open(out_DMR_counts_hmC, 'w')

    folder_out += '/data'

    out_count_mC.write('\tTotal_MR\tHyper_DMR\tHypo_DMR\n')
    out_count_hmC.write('\tTotal_MR\tHyper_DMR\tHypo_DMR\n')
    prepfiles = [line.rstrip('\n') for line in open(listfile)]

    #sorting files by size, largest to smallest
    for i in range(len(prepfiles)):
        prepfiles[i] = (prepfiles[i], os.path.getsize(prepfiles[i]))
    prepfiles.sort(key=lambda filename: filename[1], reverse=True)
    for i in range(len(prepfiles)):
        prepfiles[i] = prepfiles[i][0]

    if num_processes > 1:
        pool = mp.Pool(processes=num_processes)
        parallel = True
    else:
        parallel = False

    files = []
    if parallel:
        files = pool.map(call_fdmr, [(prepfile, test_method, fdr, cutoff, plot, folder_out) for prepfile in prepfiles], 1)
        pool.close()
    else:
        for prepfile in prepfiles:
            files.append(call_fdmr((prepfile, test_method, fdr, cutoff, plot, folder_out)))

    files = list(filter(None, files))
    for group in files:
        #group = out_filename, num_MRs, num_DMRs, num_hyper, num_hypo
        out_list_file.write(group[0]+'\n')
        count_name = group[0].split('/')[-1][:-4]

        if '5mc' in group[0].lower():
            out_count_mC.write('%s\t%d\t%d\t%d\n'%(count_name, group[1],
                                                   group[3], group[4]))
        elif '5hmc' in group[0].lower():
            out_count_hmC.write('%s\t%d\t%d\t%d\n'%(count_name, group[1],
                                                   group[3], group[4]))

    out_list_file.close()
    out_count_mC.close()
    out_count_hmC.close()


def my_parser(parser):
    required = parser.add_argument_group("Required")
    required.add_argument("-f", "--prepfile",
                        metavar='',
                        help="file list_prepared_for_DMR_finding_... containing preprocessed files, either increased or imputed")
    optional = parser.add_argument_group("Optional, has default values")
    optional.add_argument("-cp", "--cutoffPval",
                        type=float,
                        default=0.05,
                        metavar='',
                        help="the cutoff pvalue. default=0.05")
    optional.add_argument("-p", "--numProcesses",
                        default=1,
                        type=int,
                        metavar='',
                        help="enter the number of processes you want to run the filtering and formatting of extracted sites with. default=1, meaning there is no parallelization.")
    optional.add_argument("-T", "--testMethod",
                        default='Pranksum',
                        metavar='',
                        help="which test method to use. Ttest (two sampled T-test), KStest (Kolmogorov–Smirnov test), either Mranksum(Matlab ranksum), Rranksum(running R wilcox) or Pranksum(full enumeration if length of MR is less than 10, and scipy.stats.mannwhitneyu for lengths bigger than 10.), default=Pranksum")
    optional.add_argument("-fdr", "--falseDiscoveryRatio",
                        type=float,
                        default=-1,
                        metavar='',
                        help="the false discovery rate for multiple comparisons using the Benjamini/Hochberg method. if not provided, fdr won't be used. default=None")
    optional.add_argument("-pl", "--plot",
                        default='no',
                        metavar='',
                        help="if you want to plot each DMR. yes or no. default=no")
    optional.add_argument("-F", "--folderOut",
                        default='out',
                        metavar='',
                        help="what should be the name of the out-folder. default=out")
    return parser

def run(args):
    """Calls the main function with the unpacked arguments"""
    if args.folderOut[-1] == '/':
        args.folderOut = args.folderOut[:-1]

    if not os.path.exists(args.folderOut):
        os.system('mkdir '+args.folderOut)

    if not os.path.exists(args.folderOut+'/data'):
        os.system('mkdir '+args.folderOut+'/data')

    main(args.prepfile, args.cutoffPval, args.testMethod, args.falseDiscoveryRatio,
         args.plot, args.numProcesses, args.folderOut)

if __name__ == "__main__":
    args = my_parser(argparse.ArgumentParser('python prepare_for_DMR_finding.py')).parse_args()

    run(args)
