import pandas as pd
import numpy as np
import argparse
import os
import sys
from .scripts.quantile_normalization import main as qn_main
from .scripts.rawMeth_To_5mC_5hmC import main as mc_main
from .scripts.compare_5mC_5hmC_sites import main as site_main

def run_on_files(files, folder_out, quantile_norm, human, numerical_chr_name,
            bp_extension, genome_file, cutoff, type, ko_wt, sites_f_list, gzip, methylated):
    """Calls quantile_normalization if quantile_norm is True, calls rawMeth_To_5mC_5hmC
    with one sample file at a time, and calls compare_5mC_5hmC_sites, creating
    one count file per sample, counting number of significant modified sites/sites"""
    out_list = folder_out+'/list_mC_hmC_files_'+ko_wt+'.txt'
    outfile = open(out_list, 'w')

    folder_out += '/data'

    tmp_type = type
    count = 1
    for file in files:
        if tmp_type is not None:
            tmp_type += str(count)
        if not methylated:
            if quantile_norm:
                # Quantile normalization
                file = qn_main(file, gzip, folder_out)

        # Create 5mC and 5hmC BED files
        bed_files = mc_main(file, human, numerical_chr_name, bp_extension, methylated, gzip, tmp_type, folder_out)

        if methylated:
            methylated_file = bed_files[0]
            outfile.write(methylated_file+'\n')

            site_file = site_main(methylated_file, '', methylated, genome_file, cutoff, folder_out)
        else:
            _, _, _, mc_file, hmc_file = bed_files
            outfile.write(mc_file+'\n')
            outfile.write(hmc_file+'\n')

            # Create the count sites csv file
            site_file = site_main(mc_file, hmc_file, methylated, genome_file, cutoff, folder_out)
        sites_f_list.write(site_file+'\n')
        count += 1
        if tmp_type is not None:
            tmp_type = tmp_type[:-1]

    outfile.close()
    return out_list

def main(ko_files, wt_files, methylated, genome_file, human, quantile_norm, numerical_chr_name,
         bp_extension, cutoff, gzip, folder_out):
    if ko_files is None or wt_files is None:
        print('***** ERROR in data_preprocessing.py: At least one of each (normal,knockout) condition raw data file must be provided. Exiting. *****')
        exit(1)

    files = []
    for f in ko_files+wt_files:
        files.append(f.split('/')[-1])

    if len(np.unique(files)) != len((ko_files+wt_files)):
        cond1 = 'KO'
        cond2 = 'WT'
    else:
        cond1 = None
        cond2 = None

    if methylated.lower() in ('yes', 'y'):
        methylated = True
    elif methylated.lower() in ('no', 'n'):
        methylated = False
    else:
        print('***** ERROR in data_preprocessing.py: methylated must be either [yes/y] or [no/n]. Exiting. *****')
        exit(1)

    if quantile_norm.lower() in ('yes', 'y'):
        quantile_norm = True
    elif quantile_norm.lower() in ('no', 'n'):
        quantile_norm = False
    else:
        print('***** ERROR in data_preprocessing.py: quantileNorm must be either [yes/y] or [no/n]. Exiting. *****')
        exit(1)

    if gzip in ('yes', 'y'):
        gzip = True
    elif gzip in ('no', 'n'):
        gzip = False
    else:
        print('***** ERROR in data_preprocessing.py: gzip must be either [yes/y] or [no/n]. Exiting. *****')
        exit(1)

    sites_files_list = folder_out+'/list_count_sites_files.txt'
    sites_f_list = open(sites_files_list, 'w')

    out_list_files = []
    if ko_files is not None:
        out_list_file = run_on_files(ko_files, folder_out, quantile_norm, human,
                                numerical_chr_name, bp_extension, genome_file,
                                cutoff, cond1, 'KO', sites_f_list, gzip, methylated)
        out_list_files.append(out_list_file)
    if wt_files is not None:
        out_list_file = run_on_files(wt_files, folder_out, quantile_norm, human,
                                numerical_chr_name, bp_extension, genome_file,
                                cutoff, cond2, 'WT', sites_f_list, gzip, methylated)
        out_list_files.append(out_list_file)

    sites_f_list.close()

    return out_list_files

def my_parser(parser):
    important = parser.add_argument_group("Methylation data")
    important.add_argument("-fko", "--fileKO",
                        action='append',
                        metavar='file',
                        help="KNOCKOUT CONDITION raw methylation data file, one or many")
    important.add_argument("-fwt", "--fileWT",
                        action='append',
                        metavar='file',
                        help="NORMAL CONDITION raw methylation data file, one or many")
    important.add_argument("-m", "--methylated",
                        required=True,
                        metavar='y/n',
                        help="Does the methylation files contain methylation levels already? [yes/y] or [no/n]. If no it must be raw data from HMST-Seq.")
    important.add_argument("-z", "--gzip",
                        required=True,
                        metavar='y/n',
                        help="Are the methylation data in gzip format? [yes/y] or [no/n]. If no, it should be a regular txt file")

    required = parser.add_argument_group("Required")
    required.add_argument("-g", "--genomeFile",
                        required=True,
                        metavar='file',
                        help="the genome file, one column with chromosome names and one column with the length of each chromosome.")
    required.add_argument("-hu", "--human",
                        required=True,
                        metavar='y/n',
                        help="is sample from human? [yes/y] or [mouse/m] or [rat/r] for human , mouse or rat genome")
    required.add_argument("-n", "--numericalChr",
                        required=True,
                        metavar='y/n',
                        help="will you operate with numerical chromosome names or not, [yes/y] or [no/n]. the same as chosen here must be as in the genomeFile")

    optional = parser.add_argument_group("Optional, has default values")
    optional.add_argument("-qn", "--quantileNorm",
                        metavar='',
                        default='no',
                        help="run quantile normalization on raw data? [yes/y] or [no/n], default=no")
    optional.add_argument("-e", "--bpExtension",
                        type=int,
                        metavar='',
                        default=0,
                        help="extension from cg position on 5mC and 5hmC. default=0")
    optional.add_argument("-c", "--cutoff",
                        type=float,
                        default=1.,
                        metavar='',
                        help="minimum methylation level for a site to be considered methylated in the summary report. default=1")
    optional.add_argument("-F", "--folderOut",
                        default='out',
                        metavar='',
                        help="what should be the name of the out-folder. default=out")
    return parser

def run(args):
    """Calls the main function with the unpacked arguments"""
    if args.folderOut[-1] == '/':
        args.folderOut = args.folderOut[:-1]

    if not os.path.exists(args.folderOut):
        os.system('mkdir '+args.folderOut)

    if not os.path.exists(args.folderOut+'/data'):
        os.system('mkdir '+args.folderOut+'/data')

    main(args.fileKO, args.fileWT, args.methylated, args.genomeFile, args.human, args.quantileNorm,
         args.numericalChr, args.bpExtension, args.cutoff, args.gzip, args.folderOut)

if __name__ == "__main__":
    args = my_parser(argparse.ArgumentParser("python data_preprocessing.py")).parse_args()
    run(args)
