import argparse
import os
import sys
import itertools
import multiprocessing as mp
from .scripts.extract_regions_find_MRs import main as xmr_main
import time

def call_ex_mr_main(args):
    """Calls extract_regions_find_MRs with one methylation file at a time"""
    methylation_file, region_file, reference_file, min_overlap, min_consecutive,\
        adjacency,  out_folder, isWGS, cond, reg = args[0]
    start = time.time()
    MR_file = xmr_main(methylation_file, region_file, reference_file, min_overlap,
                         min_consecutive, adjacency, isWGS, reg, out_folder)
    print(MR_file+' -- '+str(time.time()-start)+' seconds')

    return MR_file, cond, reg

def run_on_meth_files(mfiles, region_files, regs, minimum_consecutive1, minimum_consecutive2,
                     minimum_consecutive3, reference_file, min_overlap, num_processes,
                     adjacency, isWGS,  folder_out):
    """Runs call_ex_mr_main with combinations of variables, depending on what kind of
    genomic region it is

    mfiles -- list of methylation files
    region_files -- filename of file containing all region files
    regs -- list of region_files
    minimum_consecutive1,2,3 -- depending on what region it is, the minimum number of consecutive sites alters
    reference_file -- reference filename
    min_overlap -- minimum overlap, fraction for bedtools intersect
    num_processes - integer bigger than or equal to 1
    adjacency -- the max number of
    isWGS -- True or False for WGS input data
    """

    combos = []
    for mfile in mfiles:
        for reg in regs:
            reg_name = reg[1]
            reg_file = reg[0]
            if reg_name in ['TSS', 'TES']:
                min_consecutive = minimum_consecutive1
            elif reg_name in ['geneBody', '5dist', 'intergenic']:
                min_consecutive = minimum_consecutive2
            elif reg_name in ['enhancer']:
                min_consecutive = minimum_consecutive3

            combos.append((mfile[0], reg_file, reference_file, min_overlap,
                           min_consecutive, adjacency, folder_out, isWGS, mfile[1], reg_name))

    if num_processes > 1:
        pool = mp.Pool(processes=num_processes)
        parallel = True
    else:
        parallel = False

    files = []
    if parallel:
        files = pool.map(call_ex_mr_main, itertools.product(combos), 1)
        pool.close()
    else:
        files = map(call_ex_mr_main, itertools.product(combos))

    return files

def main(ko_files, wt_files, reference_file, region_files, enhancer_file, min_overlap,
         num_processes, minimum_consecutive1, minimum_consecutive2,
         minimum_consecutive3, adjacency, isWGS, folder_out):

    if enhancer_file is None and region_files is None:
        print('***** ERROR in find_MRs.py: User must input regionFiles AND/OR enhancerFile. Exiting. *****')
        exit(1)

    if ko_files is None and wt_files is None:
        print('***** ERROR in find_MRs.py: User must input koFiles AND/OR wtFiles. Exiting. *****')
        exit(1)

    if ko_files is not None:
        meths_ko = [line.rstrip('\n') for line in open(ko_files)]
        for i in range(len(meths_ko)):
            meths_ko[i] = (meths_ko[i], os.path.getsize(meths_ko[i]))
        meths_ko.sort(key=lambda filename: filename[1], reverse=True)
        for i in range(len(meths_ko)):
            meths_ko[i] = meths_ko[i][0]


    if wt_files is not None:
        meths_wt = [line.rstrip('\n') for line in open(wt_files)]
        for i in range(len(meths_wt)):
            meths_wt[i] = (meths_wt[i], os.path.getsize(meths_wt[i]))
        meths_wt.sort(key=lambda filename: filename[1], reverse=True)
        for i in range(len(meths_wt)):
            meths_wt[i] = meths_wt[i][0]

    regs = []
    outfile_ko = None
    outfile_wt = None

    if enhancer_file is not None:
        regs = [(enhancer_file, 'enhancer')]
    else:
        regs = []
    if region_files is not None:
        tmp_regs = [line.rstrip('\n') for line in open(region_files)]
        for reg in tmp_regs:
            if 'tss_up' in reg.lower():
                regs.append((reg, 'TSS'))
            elif 'tes_up' in reg.lower():
                regs.append((reg, 'TES'))
            elif 'gene_up' in reg.lower():
                regs.append((reg, 'geneBody'))
            elif '5dist_up' in reg.lower():
                regs.append((reg, '5dist'))
            elif 'intergenic' in reg.lower():
                regs.append((reg, 'intergenic'))

    for i in range(len(regs)):
        regs[i] = (regs[i], os.path.getsize(regs[i][0]))
    regs.sort(key=lambda filename: filename[1], reverse=True)
    for i in range(len(regs)):
        regs[i] = regs[i][0]

    if ko_files is not None:
        out_list_file_ko = folder_out+'/list_all_filtered_formatted_MRs_KO.txt'
        outfile_ko = open(out_list_file_ko, 'w')
    if wt_files is not None:
        out_list_file_wt = folder_out+'/list_all_filtered_formatted_MRs_WT.txt'
        outfile_wt = open(out_list_file_wt, 'w')

    #if enhancer_file is not None:
    #    out_list_enhancer = folder_out+'/list_enhancer_filtered_formatted_MRs.txt'
    #    outfile_enhancer = open(out_list_enhancer, 'w')

    all_mfiles = []
    if ko_files is not None:
        for m_file in meths_ko:
            all_mfiles.append((m_file, 'ko'))

    if wt_files is not None:
        for m_file in meths_wt:
            all_mfiles.append((m_file, 'wt'))

    folder_out += '/data'

    files_ = run_on_meth_files(all_mfiles, region_files, regs, minimum_consecutive1,
                     minimum_consecutive2, minimum_consecutive3, reference_file, min_overlap,
                     num_processes, adjacency, isWGS, folder_out)

    for file, cond, reg in files_:
        if file != '':
            if cond == 'ko':
                outfile_ko.write(file+'\n')
            else:
                outfile_wt.write(file+'\n')


    out = []
    if region_files is not None:
        if ko_files is not None:
            outfile_ko.close()
            out.append(out_list_file_ko)
        if wt_files is not None:
            outfile_wt.close()
            out.append(out_list_file_wt)
    return out

def my_parser(parser):
    important = parser.add_argument_group("HMST Seq data")
    important.add_argument("-fko", "--koFiles",
                        metavar='file',
                        help="the list_mC_hmC_files_KO.txt containing the 5mC and 5hmC filenames for KNOCKOUT CONDITION samples")
    important.add_argument("-fwt", "--wtFiles",
                        metavar='file',
                        help="the list_mC_hmC_files_WT.txt containing the 5mC and 5hmC filenames for NORMAL CONDITION samples")

    required = parser.add_argument_group("Required")
    required.add_argument("-ref", "--referenceFile",
                        required=True,
                        metavar='file',
                        help="the uniquely sorted reference BED formatted file.")

    optional = parser.add_argument_group("Optional, has default values")
    optional.add_argument("-reg", "--regionFiles",
                        metavar='',
                        help="the list_region_files.txt containing the region filenames")
    optional.add_argument("-e", "--enhancerFile",
                        metavar='',
                        help="enhancer BED file")
    optional.add_argument("-f", "--minOverlap",
                        default=1E-9,
                        type=float,
                        metavar='',
                        help="the minimum overlap required as a fraction of region file, for bedtools intersect when extracting regions from methylation files. default=1E-9, (i.e. 1bp)")
    optional.add_argument("-p", "--numProcesses",
                        default=1,
                        type=int,
                        metavar='',
                        help="enter the number of processes you want to run the filtering and formatting of extracted sites with. default=1, meaning there is no parallelization.")
    optional.add_argument("-mc1", "--minimumConsecutive1",
                        default=3,
                        type=int,
                        metavar='',
                        help="the minimum number of consecutive sites for TSS and TES in a methylation region. default=3")
    optional.add_argument("-mc2", "--minimumConsecutive2",
                        default=5,
                        type=int,
                        metavar='',
                        help="the minimum number of consecutive sites for geneBody, 5distance and intergenic in a methylation region. default=5")
    optional.add_argument("-mc3", "--minimumConsecutive3",
                        default=3,
                        type=int,
                        metavar='',
                        help="the minimum number of consecutive sites for enhancer in a methylation region. default=3")
    optional.add_argument("-a", "--adjacency",
                        default=2000,
                        type=int,
                        metavar='',
                        help="the maximum number of bp between adjacent sites in a methylation region or the window bin size for WGS data when -W is Yes. default=2000.")
    # added wang for WGS
    optional.add_argument("-W", "--isWGS",
                          default='no',
                          metavar='',
                          help="the input data is WGS platform and try to use equal sized window bins to find MRs. default=no")
    optional.add_argument("-F", "--folderOut",
                        default='out',
                        metavar='',
                        help="what should be the name of the out-folder. default=out")


    return parser

def run(args):
    """Calls the main function with the unpacked arguments"""
    if args.folderOut[-1] == '/':
        args.folderOut = args.folderOut[:-1]

    if not os.path.exists(args.folderOut):
        os.system('mkdir '+args.folderOut)

    if not os.path.exists(args.folderOut+'/data'):
        os.system('mkdir '+args.folderOut+'/data')

    #added wang
    if args.isWGS.lower() in ('yes', 'y'):
        isWGS = True
    elif args.isWGS.lower() in ('no', 'n'):
        isWGS = False
    else:
        print('***** ERROR in find_MRs.py: No info whether it is WGS or not WGS type of data. If WGS data, enter [yes/y]. If other types data, enter [no/n]. Exiting. *****')
        exit(1)
    if isWGS:
        print("Analyzing data with equal sized window bins")
        print("Window bin size ", args.adjacency)
        print("Minimum number of sites in each bin ", args.minimumConsecutive1, args.minimumConsecutive2, args.minimumConsecutive3)
    else:
        print("Analyzing data with BGI MRs search method")
        print("Maximum adjaceny distance between two sites ", args.adjacency)
        print("Minimum number of sites in a MR ", args.minimumConsecutive1, args.minimumConsecutive2, args.minimumConsecutive3)

    main(args.koFiles, args.wtFiles, args.referenceFile, args.regionFiles, args.enhancerFile,
         args.minOverlap, args.numProcesses, args.minimumConsecutive1,
         args.minimumConsecutive2, args.minimumConsecutive3, args.adjacency, isWGS,
         args.folderOut)

if __name__ == "__main__":
    args = my_parser(argparse.ArgumentParser('python find_MRs.py')).parse_args()

    run(args)
