import argparse
import os
import sys
from .scripts.preprocess_MRs import main as prep_main
import multiprocessing as mp
import time

def get_methylation_and_region(filename):
    """Finds methylation(5mC or 5hmC) and genomic region based on filename and returns both"""
    if 'REGION_TSS' in filename:
        region = 'TSS'
    elif 'REGION_TES' in filename:
        region = 'TES'
    elif 'REGION_geneBody' in filename:
        region = 'geneBody'
    elif 'REGION_5dist' in filename:
        region = '5dist'
    elif 'REGION_intergenic' in filename:
        region = 'intergenic'
    elif 'REGION_enhancer' in filename:
        region = 'enhancer'
    else:
        print('***** ERROR in prepare_for_DMR_finding.py: Don\'t recognize the region type of the file: '+filename+'. The filename should not be changed. Exiting *****')
        exit(1)

    if '5mc' in filename.lower():
        methylation = '5mC'
    elif '5hmc' in filename.lower():
        methylation = '5hmC'
    else:
        print('***** ERROR in prepare_for_DMR_finding.py: Don\'t recognize the methylation type of file: '+filename+'. The filename should not be changed. Exiting *****')
        exit(1)

    return region, methylation

def call_prep_MR(args):
    """Calls preprocess_MRs with one combination - KO vs WT

    args:
      combo -- ko_mr_file, wt_mr_file, min_consecutive
      impute -- zeros/median
      adjacency -- the maximum number of bp between adjacent sites in an MR
      folder_out -- output folder name"""
    #added wang
    combo, isSameTrend, impute, adjacency, two_samples, folder_out = args

    region, methylation = get_methylation_and_region(combo[0])
    sample_ko = combo[0].split('/')[-1].split('_')[0]
    sample_wt = combo[1].split('/')[-1].split('_')[0]
    start_time = time.time()
    #added wang
    combo, single = prep_main(combo[0], combo[1], isSameTrend, impute, adjacency,
                                   combo[2], combo[3], combo[4], two_samples, folder_out)
    print(sample_ko+' vs '+sample_wt+', '+region+', '+methylation+' -- '+str(time.time()-start_time)+' seconds')

    return combo, single


def main(ko_mr_files, wt_mr_files, min_consecutive1, min_consecutive2, min_consecutive3, adjacency, isSameTrend,
         impute, num_processes, folder_out):
    """Combines same genomic regions and methylation state of KO with WT, with the
    respective min_consecutive variable"""

    if impute not in ('zeros', 'median'):
        print('***** ERROR in prepare_for_DMR_finding.py: impute must be either zeros or median. Default is zeros. Exiting *****')
        exit(1)

    ko_mrs = [line.rstrip('\n') for line in open(ko_mr_files)]
    wt_mrs = [line.rstrip('\n') for line in open(wt_mr_files)]

    #sorting files by size, largest to smallest
    for i in range(len(ko_mrs)):
        ko_mrs[i] = (ko_mrs[i], os.path.getsize(ko_mrs[i]))
    ko_mrs.sort(key=lambda filename: filename[1], reverse=True)
    for i in range(len(ko_mrs)):
        ko_mrs[i] = ko_mrs[i][0]

    for i in range(len(wt_mrs)):
        wt_mrs[i] = (wt_mrs[i], os.path.getsize(wt_mrs[i]))
    wt_mrs.sort(key=lambda filename: filename[1], reverse=True)
    for i in range(len(wt_mrs)):
        wt_mrs[i] = wt_mrs[i][0]


    out_list_file_imputed = folder_out+'/list_prepared_for_DMR_finding_imputed.txt'
    outfile_imputed = open(out_list_file_imputed, 'w')

    out_list_file_increased = folder_out+'/list_prepared_for_DMR_finding_increased.txt'
    outfile_increased = open(out_list_file_increased, 'w')

    if num_processes > 1:
        pool = mp.Pool(processes=num_processes)
        parallel = True
    else:
        parallel = False

    #checking if there are more than one of each KO and WT sample:
    kos = []
    for ko in ko_mrs:
        kos.append(ko.split('/')[-1].split('_')[0])
    num_ko_samples = len(set(kos))

    wts = []
    for wt in wt_mrs:
        wts.append(wt.split('/')[-1].split('_')[0])
    num_wt_samples = len(set(wts))

    if num_wt_samples == 1 and num_ko_samples == 1:
        two_samples = True
        out_list_file_MRs = folder_out+'/list_overlapping_MRs.txt'
        outfile_MRs = open(out_list_file_MRs, 'w')
    else:
        two_samples = False


    combinations = []
    for ko_mr_file in ko_mrs:
        ko_reg, ko_meth = get_methylation_and_region(ko_mr_file)

        for wt_mr_file in wt_mrs:
            wt_reg, wt_meth = get_methylation_and_region(wt_mr_file)
            if ko_reg == wt_reg and ko_meth == wt_meth:
                if ko_reg == 'TES' or ko_reg == 'TSS':
                    min_consecutive = min_consecutive1
                elif ko_reg == 'geneBody' or ko_reg == '5dist' or ko_reg == 'intergenic':
                    min_consecutive = min_consecutive2
                elif ko_reg == 'enhancer':
                    min_consecutive = min_consecutive3

                combinations.append([ko_mr_file, wt_mr_file, min_consecutive, ko_reg, ko_meth])

    files = []

    if parallel:
        #added wang
        files = pool.map(call_prep_MR, [(combo, isSameTrend, impute, adjacency, two_samples, folder_out+'/data') for combo in combinations], 1)
        pool.close()
    else:
        for combo in combinations:
            #added wang
            files.append(call_prep_MR((combo, isSameTrend, impute, adjacency, two_samples, folder_out+'/data')))

    for f in files:
        comb = f[0]
        single = f[1]

        if comb is not None:
            outfile_imputed.write(comb[0]+'\n')
            outfile_increased.write(comb[1]+'\n')
        if single is not None:
            if two_samples:
                outfile_MRs.write(single[0]+'\n')
                outfile_MRs.write(single[1]+'\n')



    outfile_imputed.close()
    outfile_increased.close()
    if two_samples:
        outfile_MRs.close()

    if parallel:
        pool.close()

    return out_list_file_imputed, out_list_file_increased

def my_parser(parser):
    required = parser.add_argument_group("Required")
    required.add_argument("-ko", "--KOfilesMRs",
                        required=True,
                        metavar='file',
                        help="the list file containing all KO MRs filenames: list_all_filtered_formatted_MRs_KO.txt")
    required.add_argument("-wt", "--WTfilesMRs",
                        required=True,
                        metavar='file',
                        help="the list file containing all WT MRs filenames: list_all_filtered_formatted_MRs_WT.txt")
    optional = parser.add_argument_group("Optional, has default values")
    optional.add_argument("-i", "--impute",
                        default='zeros',
                        metavar='',
                        help="what should missing values be imputed by: zeros or median. default=zeros")
    optional.add_argument("-p", "--numProcesses",
                        default=1,
                        type=int,
                        metavar='',
                        help="enter the number of processes you want to run the filtering and formatting of extracted sites with. default=1, meaning there is no parallelization.")
    optional.add_argument("-mc1", "--minimumConsecutive1",
                        default=3,
                        type=int,
                        metavar='',
                        help="the minimum number of consecutive sites for TSS and TES in a methylation region. default=3")
    optional.add_argument("-mc2", "--minimumConsecutive2",
                        default=5,
                        type=int,
                        metavar='',
                        help="the minimum number of consecutive sites for geneBody, 5distance and intergenic in a methylation region. default=5")
    optional.add_argument("-mc3", "--minimumConsecutive3",
                        default=3,
                        type=int,
                        metavar='',
                        help="the minimum number of consecutive sites for enhancer in a methylation region. default=3")
    optional.add_argument("-a", "--adjacency",
                        default=2000,
                        type=int,
                        metavar='',
                        help="the maximum number of bp between adjacent sites in a methylation region, -a and mc1,mc2,mc3 shall be consistent to the corresponding parameters in find_MRs. default=2000.")
    optional.add_argument("-F", "--folderOut",
                        default='out',
                        metavar='',
                        help="what should be the name of the out-folder. default=out")
    #added wang
    optional.add_argument("-isST", "--isSameTrend",
                          default= 0 ,
                          type = int,
                          metavar= '',
                          help="Are CpG sites the same Trend in the same MR, default=0 (not the same trend) or =1 (the same trend)")
    return parser

def run(args):
    """Calls the main function with the unpacked arguments"""
    if args.folderOut[-1] == '/':
        args.folderOut = args.folderOut[:-1]

    if not os.path.exists(args.folderOut):
        os.system('mkdir '+args.folderOut)

    if not os.path.exists(args.folderOut+'/data'):
        os.system('mkdir '+args.folderOut+'/data')

    main(args.KOfilesMRs, args.WTfilesMRs, args.minimumConsecutive1, args.minimumConsecutive2, args.minimumConsecutive3,
         args.adjacency, args.isSameTrend, args.impute, args.numProcesses, args.folderOut)

if __name__ == "__main__":
    args = my_parser(argparse.ArgumentParser('python prepare_for_DMR_finding.py')).parse_args()

    run(args)
