import pandas as pd
import numpy as np
import argparse
import os

def extract_methpoint(df):
    """Extracts some strings from the ids/names of the MRs in df, and puts them
    in separate columns """
    new_df = pd.DataFrame(columns=['names', 'meth_point', 'power'])
    s = pd.DataFrame([x.split('||') for x in list(df['names'])])
    if len(s.columns == 3): #intergenic region
        w = pd.DataFrame([x.split(':') for x in list(s[2])])
    else: #other region
        w = pd.DataFrame([x.split(':') for x in list(s[3])])
    w2 = pd.DataFrame([x.split(':') for x in list(s[0])])
    new_df['chr'] = w2[0]
    meth_point_ix = w.columns[-1]
    new_df['meth_point'] = w[meth_point_ix]
    new_df['power'] = df['power']
    tmp_names=s[0]+'||'+s[2]+'||'+w[0]
    #new_df['names'] = new_df.assign(names=s[0]+'||'+s[2]+'||'+w[0])
    new_df['names'] = tmp_names.to_list()
    new_df['num_in_group'] = +s[1]
    get_sites = pd.DataFrame([x.split(':') for x in list(s[0])])
    new_df['start_site'] = get_sites[1]
    new_df['end_site'] = get_sites[2]

    return new_df

def all_same_sample(mCs, hmCs):
    """Checks if all files of mCs and hmCs contain data from the same sample
    mCs and hmCs are dictionaries"""
    mC_files = list(mCs.values())
    hmC_files = list(hmCs.values())
    samplename = mC_files[0].split('/')[-1].split('_')[0]

    both = mC_files+hmC_files
    for f in both:
        tmp_samplename = f.split('/')[-1].split('_')[0]
        if tmp_samplename != samplename:
            return False, None
    return True, samplename

def get_counts_region(df_mC_all, df_hmC_all, cutoff):
    """Returns 6 numbers:
    num_unique_genomic_reg_mC_hmC -- the number of unique MRs
    tot_mC_in_MRs_cutoff -- the total number of methylated CG sites in MRs
                            having methylation level higher than cutoff
    tot_hmC_in_MRs_cutoff -- the total number of hydroxymethylated CG sites in MRs
                             having methylation level higher than cutoff
    total_length_of_genomic_regions -- the total length of genomic region
    tot_mC_sites_MRs_all -- the total number of methylated CG sites in MRs
    tot_hmC_sites_MRs_all -- the total number of hydroxymethylated CG sites in MRs"""

    uniq_mC = df_mC_all[['names', 'start_site', 'end_site']].drop_duplicates()
    uniq_hmC = df_hmC_all[['names', 'start_site', 'end_site']].drop_duplicates()
    uniq_both = uniq_mC.append(uniq_hmC).drop_duplicates().reset_index(drop=True)
    num_unique_genomic_reg_mC_hmC = len(uniq_both)  #1 regions

    uniq_both['diff'] = uniq_both['end_site'].astype(int)-uniq_both['start_site'].astype(int)

    total_length_of_genomic_regions = uniq_both['diff'].sum() #4 regions

    tot_mC_sites_MRs_all = len(df_mC_all)   #5 regions
    tot_hmC_sites_MRs_all = len(df_hmC_all) #6 regions

    df_mC_cutoff = df_mC_all[df_mC_all.power > cutoff]
    tot_mC_in_MRs_cutoff = len(df_mC_cutoff)  #2 regions

    df_hmC_cutoff = df_hmC_all[df_hmC_all.power > cutoff]
    tot_hmC_in_MRs_cutoff = len(df_hmC_cutoff)  #3 regions

    return num_unique_genomic_reg_mC_hmC, tot_mC_in_MRs_cutoff, tot_hmC_in_MRs_cutoff,\
            total_length_of_genomic_regions, tot_mC_sites_MRs_all, tot_hmC_sites_MRs_all

def main(tss, tes, gene_body, five_dist, intergenic, enhancer, cutoff, out_folder):
    """Create count file, containing information on genomic regions, for one sample"""
    mCs = {}
    hmCs = {}
    if tss is not None:
        for i in tss:
            if '5mC' in i:
                mCs['TSS'] = i
            elif '5hmC' in i:
                hmCs['TSS'] = i
    if tes is not None:
        for i in tes:
            if '5mC' in i:
                mCs['TES'] = i
            elif '5hmC' in i:
                hmCs['TES'] = i
    if gene_body is not None:
        for i in gene_body:
            if '5mC' in i:
                mCs['geneBody'] = i
            elif '5hmC' in i:
                hmCs['geneBody'] = i
    if five_dist is not None:
        for i in five_dist:
            if '5mC' in i:
                mCs['fiveDist'] = i
            elif '5hmC' in i:
                hmCs['fiveDist'] = i
    if intergenic is not None:
        for i in intergenic:
            if '5mC' in i:
                mCs['intergenic'] = i
            elif '5hmC' in i:
                hmCs['intergenic'] = i
    if enhancer is not None:
        for i in enhancer:
            if '5mC' in i:
                mCs['enhancer'] = i
            elif '5hmC' in i:
                hmCs['enhancer'] = i


    all_same, samplename = all_same_sample(mCs, hmCs)

    if not all_same:
        print('***** ERROR in compare_5mC_5hmC_MRs_regions.py: All given files must be from same sample. Exiting. *****')
        exit(1)

    region_file = out_folder+'/'+samplename+'_regions_counts.csv'
    f_regions = open(region_file, 'w')
    f_regions.write('num_unique_genomic_reg_tot\ttotal_mC_in_MRs_cutoff\ttotal_hmC_in_MRs_cutoff\t'+\
                    'total_length_of_genomic_regions\ttot_mC_sites_MRs_all\ttot_hmC_sites_MRs_all\n')
    total_region = [0, 0, 0, 0, 0, 0]
    for region in ('fiveDist', 'TSS', 'geneBody', 'TES', 'intergenic', 'enhancer'):
        if not region in mCs:
            continue

        df_mC_all = pd.read_csv(mCs[region], names=['names', 'power'], sep='\t')
        df_mC_all = extract_methpoint(df_mC_all)

        if len(hmCs) == 0: #Not HMST-Seq data
            df_hmC_all = pd.DataFrame(columns=['names', 'meth_point', 'power', 'chr', 'num_in_group', 'start_site', 'end_site'])
        else:
            df_hmC_all = pd.read_csv(hmCs[region], names=['names', 'power'], sep='\t')
            df_hmC_all = extract_methpoint(df_hmC_all)

        a_reg, b_reg, c_reg, d_reg, e_reg, f_reg = get_counts_region(df_mC_all, df_hmC_all, cutoff)

        f_regions.write('%s\t%d\t%d\t%d\t%d\t%d\t%d\n'%(region, a_reg, b_reg, c_reg, d_reg, e_reg, f_reg))
        total_region[0]+=a_reg
        total_region[1]+=b_reg
        total_region[2]+=c_reg
        total_region[3]+=d_reg
        total_region[4]+=e_reg
        total_region[5]+=f_reg

    f_regions.write('%s\t%d\t%d\t%d\t%d\t%d\t%d\n'%('total',
                    total_region[0], total_region[1], total_region[2],
                    total_region[3], total_region[4], total_region[5]))
    f_regions.close()

    return region_file
