import pandas as pd
import numpy as np
import argparse
import os

def get_counts_sites(df_mC, df_hmC, cutoff):
    """Returns 4 numbers:
    tot_mC_cutoff -- the total number of methylated CG sites in MRs in this chromosome
                     having methylation level higher than cutoff
    tot_hmC_cutoff -- the total number of hydroxymethylated CG sites in MRs in
                      this chromosome having methylation level higher than cutoff
    tot_mC_sites_MRs_all -- the total number of methylated CG sites in MRs in this
                            chromosome
    tot_hmC_sites_MRs_all -- the total number of hydroxymethylated CG sites in MRs
                             in this chromosome
    """

    df_mC_cutoff = df_mC[df_mC.power > cutoff]
    tot_mC_cutoff = len(df_mC_cutoff)  #1 sites

    df_hmC_cutoff = df_hmC[df_hmC.power > cutoff]
    tot_hmC_cutoff = len(df_hmC_cutoff)  #2 sites

    tot_mC_sites_MRs_all = len(df_mC)   #4 sites
    tot_hmC_sites_MRs_all = len(df_hmC) #5 sites

    return tot_mC_cutoff, tot_hmC_cutoff, tot_mC_sites_MRs_all, tot_hmC_sites_MRs_all


def main(mc_file, hmc_file, methylated, genome_file, cutoff, out_folder):
    """Create count file, containing site counts, one for each sample"""
    if methylated:
        sample = mc_file.split('/')[-1][:-4]
        df_mC = pd.read_csv(mc_file, sep='\t', names=['chr', 'start_site', 'end_site', 'power'], skiprows=1)
        df_hmC = pd.DataFrame(columns=['chr', 'start_site', 'end_site', 'power'])
    else:
        sample = mc_file.split('/')[-1].split('_')[0]
        sample2 = hmc_file.split('/')[-1].split('_')[0]
        if '5mc' not in mc_file.lower():
            print('***** ERROR in compare_5mC_5hmC_sites.py: The given value for option: sampleFileMC must be 5mC. Exiting. *****')
            exit(1)
        if '5hmc' not in hmc_file.lower():
            print('***** ERROR in compare_5mC_5hmC_sites.py: The given value for option: sampleFileHMC must be 5hmC. Exiting. *****')
            exit(1)
        if sample != sample2:
            print('***** ERROR in compare_5mC_5hmC_sites.py: The given files must be 5mC and 5hmC from the same sample. Exiting. *****')
            exit(1)

        df_mC = pd.read_csv(mc_file, sep='\t', names=['chr', 'start_site', 'end_site', 'power'], skiprows=1)
        df_hmC = pd.read_csv(hmc_file, sep='\t', names=['chr', 'start_site', 'end_site', 'power'], skiprows=1)

    out_file = out_folder+'/'+sample+'_sites_counts.csv'
    f_sites = open(out_file, 'w')
    f_sites.write('total\ttotal_mC\ttotal_hmC\ttotal_length\tall_mC_sites\tall_hmC_sites\n')

    df_chrs = pd.read_csv(genome_file, sep='\t', names=['chr', 'length'])

    #pd.options.display.max_colwidth = 100

    chrs = df_chrs['chr'].drop_duplicates().reset_index(drop=True)

    total_sites = [0, 0, 0, 0, 0]
    for chr in chrs:
        if str(chr).isdigit():
            chr = int(chr)

        a_sit, b_sit, d_sit, e_sit = get_counts_sites(df_mC[df_mC['chr'] == chr],
                                                      df_hmC[df_hmC['chr'] == chr], cutoff)

        c_sit = df_chrs[df_chrs['chr'] == chr].length.item()

        f_sites.write('%s\t%d\t%d\t%d\t%d\t%d\t%d\n'%(chr, 1, a_sit, b_sit, c_sit, d_sit, e_sit))

        total_sites[0]+=a_sit
        total_sites[1]+=b_sit
        total_sites[2]+=c_sit
        total_sites[3]+=d_sit
        total_sites[4]+=e_sit

    f_sites.write('%s\t%d\t%d\t%d\t%d\t%d\t%d\n'%('total', len(chrs),
                    total_sites[0], total_sites[1], total_sites[2],
                    total_sites[3], total_sites[4]))
    f_sites.close()

    return out_file
