import os

import numpy as np
import pandas as pd
from numba import njit


def extract(region_file, methylation_file, reference_file, region_name, min_overlap, out_folder):
    """Runs bedtools intersect on given region and methylation file,

    region_file -- region file name
    methylation_files -- methylation file name
    min_overlap -- minimum overlap required as a fraction of region file
    out_folder -- folder name of output
    reference_file -- refFlat unique sorted BED"""
    out_methylation_name = out_folder + '/' + methylation_file.split('/')[-1][:-4]
    out_region_name = region_file.split('/')[-1][:-4]

    dist5_methylation_file = out_methylation_name + '_' + 'noGenes.bed'

    if region_name == '5dist':
        # For 5distance we first remove genes(TSS, geneBody and TES) from the two methylation files
        os.system('bedtools intersect -a ' + methylation_file + ' -b ' + reference_file + \
                  ' -v > ' + dist5_methylation_file)

        out = dist5_methylation_file[:-4] + '_' + out_region_name + '_overlap' + str(min_overlap) + '.bed'

        os.system('bedtools intersect -a ' + region_file + ' -b ' + \
                  dist5_methylation_file + ' -wa -wb -f ' + str(min_overlap) + ' > ' + out)

        os.system('rm ' + dist5_methylation_file)  # removes temporary file
    else:
        out = out_methylation_name + '_' + out_region_name + '_overlap' + str(min_overlap) + '.bed'
        os.system('bedtools intersect -a ' + region_file + ' -b ' + methylation_file + \
                  ' -wa -wb -f ' + str(min_overlap) + ' > ' + out)
    return out


def filter_groups(args):
    # type: (object) -> object
    """ Loops through the groups and returns groups passing the filtering criteria.
    Also adding a column with the length of the group

    Keyword arguments:
    groups -- DataFrames to filter
    N -- length of methylation region group >= N
    d -- adjacent length within group <= d
    :rtype: object
    """
    group, N, d = args
    final = []
    tmp = []
    for index, row in group[1].iterrows():
        if len(tmp) == 0:
            tmp.append(row.values.tolist())
        else:
            if 0 < row.dpoint_start - int(tmp[-1][4]) <= d:
                tmp.append(row.values.tolist())
            else:
                if len(tmp) >= N:
                    for l in tmp:
                        l.append(len(tmp))
                    final.append(tmp)
                tmp = [row.values.tolist()]
    if len(tmp) >= N:
        for l in tmp:
            l.append(len(tmp))
        final.append(tmp)
    return np.concatenate(final) if len(final) > 0 else np.empty([0, 8])


def filter_groups4WGS(args):
    # type: (object) -> object
    """ Loops through the groups and returns groups without filtering criteria due to large sample size in WGS data.
    Also adding a column with the length of the group

    Keyword arguments:
    groups -- DataFrames to filter
    N -- length of methylation region group >= N
    d -- adjacent length within group <= d for WGS d <<2000
    This one is too slow!! Wang
    """
    group, N, d = args
    final = []
    tmp = []
    for index, row in group[1].iterrows():
        if len(tmp) == 0:
            # tmp = (row.values.tolist(),)
            tmp.append(row.values.tolist())
        else:
            if 0 < row.dpoint_start - int(tmp[-1][4]) <= d:
                # tmp = tmp + (row.values.tolist(),)
                tmp.append(row.values.tolist())
            else:
                len_tmp = len(tmp)
                # tmp2=()
                if len_tmp >= N:
                    for l in tmp:
                        l.append(len_tmp)
                        # tmp2=tmp2+ (l+[len_tmp],)

                    final.append(tmp)
                    # final = final + tmp2
                tmp = [row.values.tolist()]
                # tmp = (row.values.tolist(),)
    len_tmp = len(tmp)
    # tmp2=()
    if len_tmp >= N:
        for l in tmp:
            l.append(len_tmp)
            # tmp2=tmp2+(l+[len_tmp],)

        final.append(tmp)
        # final = final + tmp2
    return np.concatenate(final) if len(final) > 0 else np.empty([0, 8])
    # return final if len(final) > 0 else np.empty([0, 8])


# added wang
@njit
def find_idx2bin(vv, bin_edge):
    tmp_st = np.nonzero(vv >= bin_edge)
    idx = tmp_st[0][-1]
    return idx

def filter_groups4WGS2(args):
    # type: (object) -> object
    """ Loops through the groups and returns groups without filtering criteria due to large sample size in WGS data.
    Also adding a column with the length of the group

    Keyword arguments:
    groups -- DataFrames to filter
    N -- length of methylation region group >= N
    d -- adjacent length within group <= d for WGS d <<2000
    """
    group, N, d = args
    # added wang
    group_start = group[0][1]
    group_end = group[0][2]
    bin_step = d
    bin_edge = [i for i in range(group_start , group_end + 1, bin_step)]
    dd=[]
    dd = group[1].groupby(
        pd.cut(
            group[1]['dpoint_start'],
            bin_edge
        )
    )
    final3 = []
    for l in list(dd.groups):
       sub_df=[]
       indx=[]
       tmp=[]
       indx = dd.groups[l]
       len_indx=len(indx)
       if len_indx>=N :
         sub_df = group[1].loc[indx]
         len_df=len(sub_df)
         sub_df.insert(7,'num_in_group',[str(len_df)+str(l)]*len_df,True)
         final3.append(sub_df)
    if len(final3)>0:
       final3=pd.concat(final3)
       final3=final3.reset_index()
       del final3['index']
    #if False:
    #    bin_edge = np.array([i for i in range(group_start, group_end + bin_step, bin_step)])
    #    final2 = [[None]] * len(bin_edge)
    #    for index, row in group[1].iterrows():
    #       idx2bin = find_idx2bin(row.dpoint_start, bin_edge)
    #       if None in final2[idx2bin]:
    #           final2[idx2bin] = [row.values.tolist()]
    #       else:
    #           final2[idx2bin].append(row.values.tolist())
    #    final3 = []
    #    for fl in final2:
    #       len_fl = len(fl)
    #       if len_fl >= N:
    #          for l in fl:
     #           l.append(len_fl)
      #        final3.append(fl)
    #return np.concatenate(final3) if len(final3) > 0 else np.empty([0, 8])
    return final3 if len(final3)>0 else pd.DataFrame(np.empty([0,8]))

def filter_extractions(df, N, d, group_names, isWGS):
    """Grouping dataframe on genetic region and filtering out every group
    that doesn't satisfy the specifications

    Keyword arguments:
    df -- DataFrame
    N -- length of methylation region group >= N
    d -- adjacent length within group <= d
    group_names -- list of column names of df to group by
    :rtype: object"""

    groups_df = df.groupby(group_names, sort=False)

    result = []
    x = 0
    # this can be changed
    if isWGS:
       for g in groups_df:
          #added wang
          tmp=[]
          tmp=filter_groups4WGS2([g, N, d])
          if len(tmp)>0:
              result.append(tmp)
    else:
        for g in groups_df:
            #return list
            result.append(filter_groups([g,N,d]))


    names = ['chr', 'reg_start', 'reg_end', 'name', 'dpoint_start',
             'dpoint_end', 'power', 'num_in_group']
    if isWGS:
       #print "WGS data"
       if len(result)>0:
           good_groups=pd.concat(result)
       else :
           good_groups=[]
       good_df=pd.DataFrame(good_groups, columns=names)
    else:
       if len(result) > 0:
           good_groups = np.concatenate(result)
       else:
           good_groups = []
       good_df = pd.DataFrame(good_groups, columns=names)

    good_df = good_df[names]

    return good_df


def format_df_rest(df):
    """Creates a new format for the input dataframe, and returns a list with
    items like: [new_name, power]

    df -- DataFrame to reformat"""
    final = []
    for index, row in df.iterrows():
        name = row['name'].split(':')
        chr = name[0]
        orig_start = name[1]
        orig_end = name[2]
        rna_name, region_name = name[3].split('||')
        up = name[4]
        down, gene_name = name[5].split('||')
        strand = name[6]
        reg_start = name[7]
        reg_end = name[8]
        new_name = "||".join([":".join([str(chr), str(orig_start), str(orig_end), str(strand), str(rna_name)]),
                              ":".join([str(row['num_in_group']), '0']),
                              ":".join([str(region_name), str(up), str(down)]),
                              ":".join([str(gene_name), str(strand), str(reg_start),
                                        str(reg_end), str(int((int(row.dpoint_start) + int(row.dpoint_end)) / 2.))])])
        final.append([new_name, row['power']])

    return final


def convert_format_rest(df):
    """Changing the format for filtered TSS, TES, gene and 5dist with extracted
    methylation sites

    Keyword arguments:
    df -- dataframe"""

    new_format = format_df_rest(df)

    new_df = pd.DataFrame(new_format, columns=['new_name', 'power'])
    new_df = new_df[['new_name', 'power']]

    return new_df


def format_df_intergenic(df):
    """Creates a new format for the input dataframe, and returns a list with
    items like: [new_name, power]

    df -- DataFrame"""
    final = []
    for index, row in df.iterrows():
        X, Y = row['name'].split(':')[-2:]
        new_name = "||".join([":".join([str(row['chr']), str(row['reg_start']), str(row['reg_end'])]),
                              ":".join([str(row['num_in_group']), str('0')]),
                              ":".join(['intergenic', str(X), str(Y),
                                        str(int((int(row['dpoint_start']) + int(row['dpoint_end'])) / 2.))])])
        final.append([new_name, row['power']])

    return final


def convert_format_intergenic(df):
    """Changing the format for filtered intergenic regions with extracted
    methylation sites

    Keyword arguments:
    df -- DataFrame"""

    new_format = format_df_intergenic(df)

    new_df = pd.DataFrame(new_format, columns=['new_name', 'power'])
    new_df = new_df[['new_name', 'power']]

    return new_df


def format_df_enhancer(df):
    """Changing the format for filtered intergenic regions with extracted
    methylation sites

    Keyword arguments:
    df -- DataFrame[chr, reg_start, reg_end, name, dpoint_start, dpoint_end, power, num_in_group]"""
    final = []
    for index, row in df.iterrows():
        _, _, tissue = row['name'].split('|')
        new_name = "||".join([":".join([str(row['chr']), str(row['reg_start']), str(row['reg_end'])]),
                              ":".join([str(row['num_in_group']), str('0')]),
                              'enhancer',
                              ":".join(
                                  [str(tissue), str(int((int(row['dpoint_start']) + int(row['dpoint_end'])) / 2.))])])
        final.append([new_name, row['power']])

    return final


def convert_format_enhancer(df):
    """Changing the format for filtered intergenic regions with extracted
    methylation sites

    Keyword arguments:
    df -- DataFrame"""
    new_format = format_df_enhancer(df)

    new_df = pd.DataFrame(new_format, columns=['new_name', 'power'])
    new_df = new_df[['new_name', 'power']]

    return new_df


def main(methylation_file, region_file, reference_file, min_overlap,
         min_consecutive, adjacency, isWGS, reg_name, out_folder):
    # type: (object, object, object, object, object, object, object, object) -> object

    if reference_file[-4:].lower() != '.bed':
        print(
            '***** ERROR in extract_regions_find_MRs.py: The uniquely sorted reference must be a BED formatted file. Exiting. *****')
        exit(1)

    extraction_file = extract(region_file, methylation_file, reference_file, reg_name, min_overlap, out_folder)

    methylation_name = methylation_file.split('/')[-1][:-4]
    extraction_file_no_folder = extraction_file.split('/')[-1][:-4]

    if reg_name in ['TSS', 'TES', 'geneBody', '5dist']:
        group_names = ['chr', 'reg_start', 'reg_end', 'name']
        convert_format_func = convert_format_rest
    elif reg_name == 'intergenic':
        group_names = ['chr', 'reg_start', 'reg_end']
        convert_format_func = convert_format_intergenic
    elif reg_name == 'enhancer':
        group_names = ['chr', 'reg_start', 'reg_end', 'name']
        convert_format_func = convert_format_enhancer

    df = pd.read_csv(extraction_file, usecols=[0, 1, 2, 3, 5, 6, 7], sep='\t',
                     names=['chr', 'reg_start', 'reg_end', 'name',
                            'dpoint_start', 'dpoint_end', 'power'])

    os.system('rm ' + extraction_file)

    # added wanf Can change at here for new filtering in extracted regions
    out = filter_extractions(df, min_consecutive, adjacency, group_names, isWGS)

    out_filename = ''
    if len(out) > 0:
        new_df = convert_format_func(out)
        new_df = new_df[['new_name', 'power']]
        out_filename = out_folder + '/' + extraction_file_no_folder + '_REGION_' + reg_name + \
                       '_filtered_formatted_minConsecutive' + str(min_consecutive) + '.csv'
        new_df.to_csv(out_filename, sep='\t', index=False, header=None)

    return out_filename
