import argparse
import pandas as pd
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import os
import shutil
from .ranksum import ranksum as rs
#added wang
from scipy.stats import ks_2samp
from scipy.stats import ttest_ind

def calc_pvalue(x, y, eng, method):
    """Uses a test to calculate pvalue between 5mC and 5hmC, returns pvalue
    x -- x array
    y -- y array
    eng -- matlab engine if Mranksum is the method, None else
    method -- Mranksum, Pranksum or Rranksum"""

    if method == 'Pranksum':
        p = rs(x, y, 'Pranksum')
    elif method == 'Rranksum':
        p = rs(x, y, 'R_wilcox')
    elif method == 'Mranksum':
        p = rs(x, y, 'M_ranksum', eng)
    elif method == 'KStest':
        #added wang
        ks_test=ks_2samp(x,y)
        p = ks_test[1]
    elif method == 'Ttest':
        t_test=ttest_ind(x,y)
        p= t_test[1]

    return p

def fdr_bh(pvals, sorted, alpha):
    """Performing the Benjamini-Hochberg procedure on the p-values. Returns a list
    of booleans representing which p-values to be rejected and not"""
    sort_df = pd.DataFrame({'pvals': pvals})
    sort_df.sort_values('pvals', inplace=True) #sort
    sort_df['rank'] = sort_df['pvals'].rank(method='dense') #rank
    sort_df['imQ'] = (sort_df['rank']/len(pvals))*alpha #calc i/m*alpha
    max_pval = sort_df[sort_df.pvals < sort_df.imQ].pvals.max() #find biggest pval less than its imQ
    sort_df['bool'] = sort_df['pvals'] <= max_pval #get true for every rejection
    sort_df.sort_index(inplace=True) #sort to original order

    return sort_df['bool'].values



def compare(df, method, fdr, cutoff_pval, eng, plot, plot_folders):
    """df -- ['name', 'meth_points_KO', 'num_in_group_KO', 'orig_power_KOs',
              'new_power_KOs' 'meth_points_WT', 'num_in_group_WT', 'orig_power_WTs',
              'new_power_WTs', 'median_KO', 'median_WT', 'rratio']"""

    ps = []
    for i, row in df.iterrows():
        p = calc_pvalue(row.new_power_KOs, row.new_power_WTs, eng, method)
        ps.append(p)
    df['pvals'] = ps

    rejected = df[df.pvals < cutoff_pval]
    
    #jbw
    if fdr != None:
        # Does a false discovery rate correction of the pvalues, using the
        # Benjamini/Hochberg method
        print('#rejected without FDR:', len(rejected))
        reject = fdr_bh(df.pvals.values, False, fdr)
        rejected = df[reject]

        print('#rejected after FDR:', len(rejected))

    if plot:
        plot_this(rejected, plot_folders[0])

    return rejected


def plot_this(df, plot_folder):
    """Plots DMRs/DhMRs

    df -- ['name', 'meth_points_KO', 'num_in_group_KO', 'orig_power_KOs',
           'new_power_KOs', 'meth_points_WT', 'num_in_group_WT', 'orig_power_WTs',
           'new_power_WTs', 'median_KO', 'median_WT', 'rratio']
    plot_folder -- folder name"""
    #print(df); exit(1)

    if os.path.exists(plot_folder):
        shutil.rmtree(plot_folder)
    os.system('mkdir '+plot_folder)

    for index, row in df.iterrows():
        x = np.arange(len(row.new_power_KOs))
        x2 = np.linspace(0, max(x), len(row.orig_power_KOs))
        plt.plot(x, row.new_power_WTs, '-o', label='WT imputed')
        plt.plot(x2, row.orig_power_WTs, 'D', label='WT original')
        plt.plot(x, row.new_power_KOs, '-o', label='KO imputed')
        plt.plot(x2, row.orig_power_KOs, 'D', label='KO original')
        plt.xlabel('Genome sites')
        plt.ylabel('Methylation Levels')
        bottom, top = plt.ylim()  # return the current ylim
        plt.ylim(top=top+0.25)
        plt.xticks([])

        plt.title(row['name'].split('||')[0]+'\n'+\
                  'p-value: '+str(row.pvals))
        #added Amna
        lgd= plt.legend(loc='center right', bbox_to_anchor=(1.3,0.5), ncol=1 , fancybox=True, shadow=True)
        plt.savefig(plot_folder+str(index)+'plot_len_group_'+str(len(x))+'.png',bbox_extra_artists=(lgd,), bbox_inches='tight')
        plt.close()

    return

def make_names(df):
    """Creates IDs for both KO and WT"""

    name = df['name'].str.split('\|\|')
    name0 = name.map(lambda x: x[0])
    name1_ = name.map(lambda x: '||'.join(x[1:]))
    name_KO = name0+'||'+df['num_in_group_KO']+'||'+name1_
    name_WT = name0+'||'+df['num_in_group_WT']+'||'+name1_

    #removes [, ] and weird newlines inside string
    meth_points_KO_str = df['meth_points_KO'].str.replace(r'[\[\]\n]', '')
    meth_points_WT_str = df['meth_points_WT'].str.replace(r'[\[\]\n]', '')
    name_KO += '||'+meth_points_KO_str+'||'
    name_WT += '||'+meth_points_WT_str+'||'

    return name_WT, name_KO

def str_list_toArray(str):
    """Helper function that turns a string of an array into an array of floats"""
    str = str.replace('[', '').replace(']', '')
    str_split = str.split(',')
    return np.array(str_split).astype(np.float)

def get_genename(df):
    """Adds a column to df with the gene name"""
    if len(df) == 0:
        df['gene_name'] = ''
    else:
        s = pd.DataFrame([x.split('||') for x in list(df.name_comb)])
        w = pd.DataFrame([x.split(':') for x in list(s[3])])
        df['gene_name'] = w[[0]].set_index(df.index)
    return df

def main(prepfile, test_method, fdr, cutoff, plot, out_folder):
    """Performs a comparison test on overlapping MRs in one prepared file"""

    if plot.lower() in ('yes', 'y', 'true'):
        plot = True
    else:
        plot = False
    if fdr == -1:
        fdr = None

    if '_TSS_' in prepfile:
        region = 'TSS'
    elif '_TES_' in prepfile:
        region = 'TES'
    elif '_geneBody_' in prepfile:
        region = 'geneBody'
    elif '_intergenic_' in prepfile:
        region = 'intergenic'
    elif '_5dist_' in prepfile:
        region = '5dist'
    elif '_enhancer_' in prepfile:
        region = 'enhancer'

    if '_5mC_' in prepfile:
        meth = '5mC'
    elif '5hmC' in prepfile:
        meth = '5hmC'

    eng = None
    if test_method == 'Mranksum':
        import matlab.engine
        eng = matlab.engine.start_matlab()

    samplenames = '_'.join(prepfile.split('/')[-1].split('_')[:3])
    filename = out_folder+'/'+prepfile.split('/')[-1][:-4]

    conv = {'orig_power_KOs': str_list_toArray,
            'orig_power_WTs': str_list_toArray,
            'new_power_KOs': str_list_toArray,
            'new_power_WTs': str_list_toArray}
    df = pd.read_csv(prepfile, index_col=False, header=0, converters=conv,
                       names=('name', 'meth_points_KO', 'num_in_group_KO', 'orig_power_KOs',
                              'new_power_KOs', 'meth_points_WT', 'num_in_group_WT', 'orig_power_WTs',
                              'new_power_WTs', 'median_KO', 'median_WT', 'rratio'))
    num_MRs = len(df)

    less_folder = ''; more_folder = ''
    if plot:
        less_folder = out_folder+'/'+samplenames+'_plots_DMRs_'+meth+'_'+region+'_fdr'+str(fdr)+'_testMethod_'+test_method+'/'
        more_folder = out_folder+'/'+samplenames+'_plots_NOPASS'+str(fdr)+'_testMethod_'+test_method+'/'

    DMRs = compare(df=df, method=test_method, fdr=fdr, cutoff_pval=cutoff, eng=eng,
                   plot=plot, plot_folders=(less_folder, more_folder))
    if test_method == 'Mranksum':
        eng.quit()

    name_WT, name_KO = make_names(DMRs)
    name_both = name_WT + samplenames + '||' + name_KO
    out_df = pd.DataFrame({'name_comb': name_both, 'median_WT': DMRs.median_WT,
                           'median_KO': DMRs.median_KO, 'rratio': DMRs.rratio,
                           'pvals': DMRs.pvals})

    num_DMRs = len(out_df)
    if '_intergenic_' in filename:
        out_df['gene_name'] = '.'
    else:
        out_df = get_genename(out_df)



    out_df = out_df[['name_comb', 'median_WT', 'median_KO', 'rratio', 'pvals', 'gene_name']]
    out_filename = filename[:-15]

    hyper_df = out_df[out_df['rratio'] > 0]
    hypo_df = out_df[out_df['rratio'] < 0]
    uncertain_df = out_df[out_df['rratio'] == 0] #needed?

    num_hyper = len(hyper_df)
    num_hypo = len(hypo_df)

    if len(hyper_df) > 0:
        hyper_df.drop(['gene_name'], axis=1).to_csv(out_filename+'DMRs_hyper.csv', index=False)
        if '_intergenic_' not in out_filename:
            hyper_df.to_csv(out_filename+'DMRs_hyper_geneNames.csv', index=False, columns=['gene_name'])
    if len(hypo_df) > 0:
        hypo_df.drop(['gene_name'], axis=1).to_csv(out_filename+'DMRs_hypo.csv', index=False)
        if '_intergenic_' not in out_filename:
            hypo_df.to_csv(out_filename+'DMRs_hypo_geneNames.csv', index=False, columns=['gene_name'])
    if len(uncertain_df) > 0:
        uncertain_df.drop(['gene_name'], axis=1).to_csv(out_filename+'DMRs_rratio0.csv', index=False)
        if '_intergenic_' not in out_filename:
            uncertain_df.to_csv(out_filename+'DMRs_rratio0_geneNames.csv', index=False, columns=['gene_name'])

    out_df.drop(['gene_name'], axis=1).to_csv(out_filename+'DMRs_all.csv', index=False)

    return out_filename+'DMRs_all.csv', num_MRs, num_DMRs, num_hyper, num_hypo
