import argparse
import pandas as pd
import numpy as np
import scipy.interpolate as sc_i
import seaborn as sns
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import os
from scipy.ndimage import gaussian_filter1d
import datetime
from .plot_TSSgeneTES import str_list_toArray as str_list_toArray
from .plot_TSSgeneTES import interpolater as interpolater
from .plot_TSSgeneTES import find_average as find_average


def get_methylation_type(files):
    """Finds what is the methylation type for each of the files, returns it. If
    it's not the same for all files, return None"""
    all_5mC = []
    all_5hmC = []
    for file in files:
        all_5mC.append('5mC' in file)
        all_5hmC.append('5hmC' in file)
    if all(all_5mC):
        methylation_type = '5mC'
    elif all(all_5hmC):
        methylation_type = '5hmC'
    else:
        methylation_type = None
    return methylation_type

def main(files, tissue, x, y, step_size, window, sigma, MRs_txt, folder_out):
    """Plotting the distribution of 5mC/5hmC levels across the sites of enhancer MRs"""
    # Checking if all files given are all mC or hmC
    methylation_type = get_methylation_type(files)
    if methylation_type is None:
        print('***** ERROR in plot_enhancer.py: All files must be of same methylation type, either 5mC or 5hmC. Exiting. *****')
        exit(1)

    conv = {'meth_points_all': str_list_toArray,
            'powers': str_list_toArray}

    enhancer_df = pd.DataFrame()

    for file in files:
        sample = (file.split('/')[-1]).split('_')[0]
        tmp_df = pd.read_csv(file, index_col=False, header=0, converters=conv)
        #jbw
        if tissue != 'all':
            tmp_df = tmp_df[tmp_df['tissue'] == tissue]
            if len(tmp_df) == 0:
                print('***** ERROR: The tissue type given is not found in the enhancer file. Exiting. *****')
                exit(1)
        tmp_df = interpolater(tmp_df, -x, y, step_size)
        tmp_avg = find_average(tmp_df, -x, y+1)
        enhancer_df[sample] = tmp_avg#[-x:y]

    out_df = pd.DataFrame()

    plt.rcParams.update({'font.size': 12})
    #sns.set(font_scale=1.5)
    plt.figure(figsize=(8,8))
    for sample in enhancer_df.columns:
        smooth = enhancer_df[sample].rolling(window=window, min_periods=1, center=True).mean()
        smooth = gaussian_filter1d(smooth, sigma=sigma)
        out_df[sample] = smooth

        plt.plot(enhancer_df.index, smooth, label=sample)

    out_df.to_csv(folder_out+'/plotData/plotData_enhancer_smoothed'+methylation_type+'_'+MRs_txt+'.csv')

    plt.xlim(enhancer_df.index[0], enhancer_df.index[-1])
    #jbw
    if tissue != 'all':
        plt.title(methylation_type+', tissue: '+tissue+', '+MRs_txt)
    else:
        plt.title(methylation_type+', '+MRs_txt)
    plt.ylabel('Average Methylation Level')
    plt.xticks(np.arange(enhancer_df.index[0], enhancer_df.index[-1]+1, 1000))
    plt.legend()
    plt.savefig(folder_out+'/'+methylation_type+'_enhancer_tissue'+tissue.upper()+'_X'+str(x)+'_Y'+str(y)+\
                '_binsize'+str(step_size)+'_sigma'+str(sigma)+'_'+MRs_txt+'_'+\
                str(datetime.datetime.now())[:10]+'.png')
