import pandas as pd
pd.options.mode.chained_assignment = None
import argparse
import numpy as np
import sys
np.set_printoptions(threshold=sys.maxsize)
import scipy.interpolate as sc_i
import os

def remove_len_groups_and_methpoint(df, region):
    """Exctracts the length of the groups and methylation datapoints,
    for finding unique ids, adds these as own columns. Returns new DataFrame

    df -- DataFrame on form: [names, power]
    region -- region name"""


    new_df = pd.DataFrame(columns=['names', 'meth_point', 'power', 'num_in_group'])
    s = pd.DataFrame([x.split('||') for x in list(df.names)])
    if region == 'intergenic':
        new_df = new_df.assign(names=s[0]+'||intergenic')
        meth_points = pd.DataFrame(([x.split(':')[-1] for x in list(s[2])]))
        new_df['meth_point'] = meth_points
    elif region == 'enhancer':
        wm = pd.DataFrame([x.split(':') for x in list(s[3])])
        new_df = new_df.assign(names=s[0]+'||'+s[2]+'||'+wm[0])
        new_df['meth_point'] = wm[1]
    else:
        wm = pd.DataFrame([x.split(':') for x in list(s[3])])
        new_df = new_df.assign(names=s[0]+'||'+s[2]+'||'+wm[0]+':'+wm[1]+':'+wm[2]+':'+wm[3])
        new_df['meth_point'] = wm[4]
    new_df['power'] = df['power']
    new_df['num_in_group'] = s[1]

    return new_df

def impute_missing(powers_df, method):
    """Replaces missing values of MRs methylation levels
    powers_df -- dataframe column with methylation levels
    method -- zeros, median or interpolate"""
    if method == 'zeros':
        powers_df = powers_df.fillna(0)
    elif method == 'median':
        powers_df = powers_df.fillna(powers_df.median())
    elif method == 'interpolate':
        # if the first element is NaN, interpolation doesn't work, so the simplest
        # way is to just set the first element to be the same as next non-nan element
        if np.isnan(powers_df.values[0]):
            powers_df.loc[0] = powers_df[~powers_df.isnull()].reset_index(drop=True)[0]
        powers_df = powers_df.interpolate()
    return powers_df



def group_to_row(df, min_consecutive, degree_spline=2, impute_method='zeros'):
    """Takes the methylation region dataframe(df) and turns the information contained
    into a list with [name, meth_points_KO, num_in_group_KO, orig_power_KOs,
    imputed_power_KOs, increased_power_KOs, meth_points_WT, num_in_group_WT,
    orig_power_WTs, imputed_power_WTs, increased_power_WTs, median_KO, median_WT, rratio]

    df -- DataFrame with MR
    min_consecutive -- minimum number of consecutive sites to be MR
    degree_spline -- degree of the smoothing spline, default 2
    impute_method -- zeros or median, default zeros"""

    if len(df) < min_consecutive :
        return

    if df.num_in_group_y.isnull().all():
        return
    if df.num_in_group_x.isnull().all():
        return
    df = df.reset_index(drop=True)

    interpolated_WT = df.power_WT.copy()
    interpolated_KO = df.power_KO.copy()

    imputed_WT = df.power_WT.copy()
    imputed_KO = df.power_KO.copy()

    if np.isnan(df.power_WT).any():
        interpolated_WT = impute_missing(interpolated_WT, 'zeros')
        imputed_WT = impute_missing(imputed_WT, method=impute_method)
    if np.isnan(df.power_KO).any():
        interpolated_KO = impute_missing(interpolated_KO, 'zeros')
        imputed_KO = impute_missing(imputed_KO, method=impute_method)

    x = np.arange(len(df))
    new_x = np.linspace(x.min(), x.max(), len(df)*2)
    #added wang
    if len(x)>degree_spline:
       increased_WT = sc_i.InterpolatedUnivariateSpline(x, interpolated_WT, k=degree_spline)(new_x)
       increased_KO = sc_i.InterpolatedUnivariateSpline(x, interpolated_KO, k=degree_spline)(new_x)
    else:
        increased_WT=interpolated_WT.copy(deep=True)
        increased_KO=interpolated_KO.copy(deep=True)

    #The two conditions without the imputed/zero values
    df_KO_wo_NaN = df[~df.num_in_group_x.isnull()]
    df_WT_wo_NaN = df[~df.num_in_group_y.isnull()]



    new_df = []
    new_df.append(df.names[0])                        #name

    meth_points_KO = df_KO_wo_NaN.meth_point.astype(int).tolist()
    new_df.append(meth_points_KO)                     #meth_points_KO

    num_KO = (df_KO_wo_NaN.reset_index(drop=True).num_in_group_x)[0]
    new_df.append(num_KO)                             #num_in_group_KO

    new_df.append(df.power_KO.tolist())               #orig_power_KOs
    new_df.append(list(imputed_KO))                   #imputed_power_KOs
    new_df.append(list(increased_KO))                 #increased_power_KOs

    meth_points_WT = df_WT_wo_NaN.meth_point.astype(int).tolist()
    new_df.append(meth_points_WT)                     #meth_points_WT

    num_WT = (df_WT_wo_NaN.reset_index(drop=True).num_in_group_y)[0]
    new_df.append(num_WT)                             #num_in_group_WT

    new_df.append(df.power_WT.tolist())               #orig_power_WTs
    new_df.append(list(imputed_WT))                   #imputed_power_WTs
    new_df.append(list(increased_WT))                 #increased_power_WTs

    #Median and rratio not including missing values
    median_KO = df_KO_wo_NaN.power_KO.median(skipna=True)+1
    median_WT = df_KO_wo_NaN.power_WT.median(skipna=True)+1

    rratio = float((median_KO - median_WT))/float(np.mean([median_KO, median_WT]))

    new_df.append(median_KO)                          #median_KO
    new_df.append(median_WT)                          #median_WT
    new_df.append(rratio)                             #rratio

    return new_df


def data_handler(KO_df, WT_df, min_consecutive, adjacency, impute_method, region):
    """Finds unique ids of knockout(KO) condition sample, extracts data from both KO and
    normal condition sample(WT) and outer joins on methylation data point site. Then
    it goes on to methylation regions(MRs) in the merged(KO-WT) DataFrame.
    KO_df -- DataFrame containing MRs of knockout condition sample
    WT_df -- DataFrame containing MRs of normal condition sample
    min_consecutive -- minimum number of consecutive sites in a MR
    adjacency -- the maximum number of bp between adjacent sites in a methylation region
    impute_method -- method to impute missing values
    region -- region name
    """
    KO_df = remove_len_groups_and_methpoint(KO_df, region)
    WT_df = remove_len_groups_and_methpoint(WT_df, region)

    unique_ids_1 = KO_df['names'].drop_duplicates().reset_index(drop=True)
    extracted1 = KO_df.loc[KO_df['names'].isin(unique_ids_1)]
    extracted2 = WT_df.loc[WT_df['names'].isin(unique_ids_1)]

    merged = pd.merge(extracted1, extracted2, on=['names', 'meth_point'], how='outer')

    #power_KO = methylation level of knock out condition
    #power_WT = methylation level of wide type / normal condition
    merged = merged.rename(columns={'power_x':'power_KO', 'power_y':'power_WT'})

    groups = merged.groupby('names', sort=False)

    # Removing regions where there is no WT value at all
    filtered_df = groups.filter(lambda x: ~x.num_in_group_y.isnull().all())
    filtered_df = filtered_df.reset_index(drop=True)

    groups = filtered_df.groupby('names', sort=False)

    out_df_lst = []
    for group in groups:
        group = group[1].sort_values('meth_point').reset_index(drop=True)

        #Finds length of gaps between methylation points
        group['diff'] = group.meth_point.astype(int).diff()

        # puts indexes where gap > adjacency(2000 by default) into a list
        splits = group.index[group['diff'] > adjacency].tolist()

        len_splits = len(splits)
        start = 0
        if len_splits > 0:
            for i,split in enumerate(splits):
                ret_val = group_to_row(df=group.loc[start:split-1],
                                       min_consecutive=min_consecutive,
                                       impute_method=impute_method)
                if ret_val is not None:
                    out_df_lst.append(ret_val)
                if i == len_splits-1:
                    ret_val = group_to_row(df=group.loc[split:],
                                           min_consecutive=min_consecutive,
                                           impute_method=impute_method)
                    if ret_val is not None:
                        out_df_lst.append(ret_val)
                else:
                    start = split

        else:
            ret_val = group_to_row(df=group, min_consecutive=min_consecutive)
            if ret_val is not None:
                out_df_lst.append(ret_val)

    out_df_lst = pd.DataFrame(out_df_lst,
                 columns= ['name', 'meth_points_KO', 'num_in_group_KO', 'orig_power_KOs',
                           'imputed_power_KOs', 'increased_power_KOs', 'meth_points_WT',
                           'num_in_group_WT', 'orig_power_WTs', 'imputed_power_WTs',
                           'increased_power_WTs', 'median_KO', 'median_WT', 'rratio'])\
                           .reset_index(drop=True)
    return out_df_lst


def data_handler4WGS(KO_df, WT_df, min_consecutive, adjacency, impute_method, region):
    """Finds unique ids of knockout(KO) condition sample, extracts data from both KO and
    normal condition sample(WT) and outer joins on methylation data point site. Then
    it goes on to methylation regions(MRs) in the merged(KO-WT) DataFrame.
    KO_df -- DataFrame containing MRs of knockout condition sample
    WT_df -- DataFrame containing MRs of normal condition sample
    Here is for WGS data which does not consdider min_consecutive and adjacency sites in MR
    min_consecutive -- minimum number of consecutive sites in a MR
    adjacency -- the maximum number of bp between adjacent sites in a methylation region
    impute_method -- method to impute missing values
    region -- region name
    """
    #added wang
    KO_df = remove_len_groups_and_methpoint(KO_df, region)
    WT_df = remove_len_groups_and_methpoint(WT_df, region)

    unique_ids_1 = KO_df['names'].drop_duplicates().reset_index(drop=True)
    extracted1 = KO_df.loc[KO_df['names'].isin(unique_ids_1)]
    extracted2 = WT_df.loc[WT_df['names'].isin(unique_ids_1)]

    merged = pd.merge(extracted1, extracted2, on=['names', 'meth_point'], how='outer')

    #power_KO = methylation level of knock out condition
    #power_WT = methylation level of wide type / normal condition
    merged = merged.rename(columns={'power_x':'power_KO', 'power_y':'power_WT'})

    groups = merged.groupby('names', sort=False)

    # Removing regions where there is no WT value at all
    filtered_df = groups.filter(lambda x: ~x.num_in_group_y.isnull().all())
    filtered_df = filtered_df.reset_index(drop=True)

    groups = filtered_df.groupby('names', sort=False)

    out_df_lst = []
    for group in groups:
        #added wang
        group = group[1].sort_values('meth_point').reset_index(drop=True)

        #Finds length of gaps between methylation points
        #added wang
        group['diff'] = group.meth_point.astype(int).diff()

        # puts indexes where gap > adjacency(2000 by default) into a list #wang here can put DMR with the same trend??
        splits = group.index[group['diff'] > adjacency].tolist()

        #added wang
        len_splits = len(splits)
        start = 0
        if len_splits > 0:
            for i,split in enumerate(splits):
                ret_val = group_to_row(df=group.loc[start:split-1],
                                       min_consecutive=min_consecutive,
                                       impute_method=impute_method)
                if ret_val is not None:
                    out_df_lst.append(ret_val)
                if i == len_splits-1:
                    ret_val = group_to_row(df=group.loc[split:],
                                           min_consecutive=min_consecutive,
                                           impute_method=impute_method)
                    if ret_val is not None:
                        out_df_lst.append(ret_val)
                else:
                    start = split

        else:
            ret_val = group_to_row(df=group, min_consecutive=min_consecutive)
            if ret_val is not None:
                out_df_lst.append(ret_val)

    out_df_lst = pd.DataFrame(out_df_lst,
                 columns= ['name', 'meth_points_KO', 'num_in_group_KO', 'orig_power_KOs',
                           'imputed_power_KOs', 'increased_power_KOs', 'meth_points_WT',
                           'num_in_group_WT', 'orig_power_WTs', 'imputed_power_WTs',
                           'increased_power_WTs', 'median_KO', 'median_WT', 'rratio'])\
                           .reset_index(drop=True)
    return out_df_lst

def data_handler4WGS_and_pattern(KO_df, WT_df, min_consecutive, adjacency, impute_method, region):
    """Finds unique ids of knockout(KO) condition sample, extracts data from both KO and
    normal condition sample(WT) and outer joins on methylation data point site. Then
    it goes on to methylation regions(MRs) in the merged(KO-WT) DataFrame.
    KO_df -- DataFrame containing MRs of knockout condition sample
    WT_df -- DataFrame containing MRs of normal condition sample
    Here is for WGS data which does not consdider min_consecutive and adjacency sites in MR
    min_consecutive -- minimum number of consecutive sites in a MR
    adjacency -- the maximum number of bp between adjacent sites in a methylation region
    impute_method -- method to impute missing values
    region -- region name
    """
    #added wang aug 2020 for separate pattern with similar trend in DMR
    KO_df = remove_len_groups_and_methpoint(KO_df, region)
    WT_df = remove_len_groups_and_methpoint(WT_df, region)

    unique_ids_1 = KO_df['names'].drop_duplicates().reset_index(drop=True)
    extracted1 = KO_df.loc[KO_df['names'].isin(unique_ids_1)]
    extracted2 = WT_df.loc[WT_df['names'].isin(unique_ids_1)]

    merged = pd.merge(extracted1, extracted2, on=['names', 'meth_point'], how='outer')

    #power_KO = methylation level of knock out condition
    #power_WT = methylation level of wide type / normal condition
    merged = merged.rename(columns={'power_x':'power_KO', 'power_y':'power_WT'})

    groups = merged.groupby('names', sort=False)

    # Removing regions where there is no WT value at all
    filtered_df = groups.filter(lambda x: ~x.num_in_group_y.isnull().all())
    filtered_df = filtered_df.reset_index(drop=True)

    groups = filtered_df.groupby('names', sort=False)

    out_df_lst = []
    for group in groups:
        #added wang
        group = group[1].sort_values('meth_point').reset_index(drop=True)

        #Finds length of gaps between methylation points
        #added wang
        group['diff'] = group.meth_point.astype(int).diff()
        tmp_KO=group.power_KO.astype(float)
        tmp_WT=group.power_WT.astype(float)
        tmp_KO=impute_missing(tmp_KO, 'zeros')
        tmp_WT=impute_missing(tmp_WT, 'zeros')
        group['delt'] = tmp_KO - tmp_WT

        # puts indexes where gap > adjacency(2000 by default) into a list #wang here can put DMR with the same trend??
        splits = group.index[group['diff'] > adjacency].tolist()

        #added wang
        len_splits = len(splits)
        start = 0
        if len_splits > 0:
            for i,split in enumerate(splits):
                #added wang
                tmp_df=group.loc[start:split-1]
                tmp_df_pos=tmp_df[tmp_df.delt>=0]
                tmp_df_neg=tmp_df[tmp_df.delt<0]
                #ret_val = group_to_row(df=group.loc[start:split-1],
                #                       min_consecutive=min_consecutive,
                #                       impute_method=impute_method)
                ret_val_pos= group_to_row(df=tmp_df_pos, min_consecutive=min_consecutive, impute_method=impute_method)
                ret_val_neg= group_to_row(df=tmp_df_neg, min_consecutive=min_consecutive, impute_method=impute_method)
                if ret_val_pos is not None:
                    out_df_lst.append(ret_val_pos)
                if ret_val_neg is not None:
                    out_df_lst.append(ret_val_neg)

                if i == len_splits-1:
                    tmp_df=group.loc[split:]
                    tmp_df_pos=tmp_df[tmp_df.delt>=0]
                    tmp_df_neg=tmp_df[tmp_df.delt<0]
                    #ret_val = group_to_row(df=group.loc[split:],
                    #                       min_consecutive=min_consecutive,
                    #                       impute_method=impute_method)
                    ret_val_pos = group_to_row(df=tmp_df_pos, min_consecutive=min_consecutive,impute_method=impute_method)
                    ret_val_neg = group_to_row(df=tmp_df_neg, min_consecutive=min_consecutive, impute_method=impute_method)

                    if ret_val_pos is not None:
                        out_df_lst.append(ret_val_pos)
                    if ret_val_neg is not None:
                        out_df_lst.append(ret_val_neg)
                else:
                    start = split

        else:
            tmp_df=group
            tmp_df_neg=tmp_df[tmp_df.delt<0]
            tmp_df_pos=tmp_df[tmp_df.delt>=0]

            #ret_val = group_to_row(df=group, min_consecutive=min_consecutive)
            ret_val_pos = group_to_row(df=tmp_df_pos, min_consecutive=min_consecutive)
            ret_val_neg = group_to_row(df=tmp_df_neg, min_consecutive=min_consecutive)
            if ret_val_pos is not None:
                out_df_lst.append(ret_val_pos)
            if ret_val_neg is not None:
                out_df_lst.append(ret_val_neg)

    out_df_lst = pd.DataFrame(out_df_lst,
                 columns= ['name', 'meth_points_KO', 'num_in_group_KO', 'orig_power_KOs',
                           'imputed_power_KOs', 'increased_power_KOs', 'meth_points_WT',
                           'num_in_group_WT', 'orig_power_WTs', 'imputed_power_WTs',
                           'increased_power_WTs', 'median_KO', 'median_WT', 'rratio'])\
                           .reset_index(drop=True)
    return out_df_lst


def clear_for_plot(df):
    """region start, region end, meth_points_all, powers"""
    pd.set_option('display.max_colwidth', 90)
    KO = pd.DataFrame(columns=['region_start', 'region_end', 'meth_points_all', 'powers'])
    WT = pd.DataFrame(columns=['region_start', 'region_end', 'meth_points_all', 'powers'])
    s = pd.DataFrame([x.split(':') for x in list(df.name)])
    if len(s) == 0:
        return None, None
    s2_list = pd.DataFrame([x.split('||') for x in list(s[2])])

    KO = KO.assign(region_start=s[1])
    KO = KO.assign(region_end=s2_list[0])
    KO['meth_points_all'] = df.meth_points_KO
    KO = KO.assign(powers=[list(p for p in power if str(p) != 'nan') for power in df.orig_power_KOs])

    WT = WT.assign(region_start=s[1])
    WT = WT.assign(region_end=s2_list[0])
    WT['meth_points_all'] = df.meth_points_WT
    WT = WT.assign(powers=[list(p for p in power if str(p) != 'nan') for power in df.orig_power_WTs])

    if len(s2_list.iloc[0]) > 1 and s2_list.iloc[0][1] == 'enhancer':
        WT['tissue'] = s2_list[2]
        KO['tissue'] = s2_list[2]
    else:
        WT['tissue'] = '-'
        KO['tissue'] = '-'

    return KO, WT


def main(KO_file, WT_file, isSameTrend, impute, adjacency, min_consecutive, region, methylation, two_samples, out_folder):
    """Finds overlapping MRs between the KO_file and WT_file samples"""

    KO_samplename = KO_file.split('/')[-1].split('_')[0]
    WT_samplename = WT_file.split('/')[-1].split('_')[0]

    samplenames = KO_samplename+'_vs_'+WT_samplename
    filename = samplenames+'_'+methylation+'_'+region

    KO_df = pd.read_csv(KO_file, sep='\t', names=('names', 'power'))
    WT_df = pd.read_csv(WT_file, sep='\t', names=('names', 'power'))

    #added WGS and pattern by wang aug 2020
    if isSameTrend ==1:
       out_df = data_handler4WGS_and_pattern(KO_df,
                          WT_df,
                          min_consecutive,
                          adjacency,
                          region=region,
                          impute_method=impute)
       print("CpG sites with the same trend in the same MR")
    elif isSameTrend ==0:
        out_df= data_handler4WGS(KO_df,
                                 WT_df,
                                 min_consecutive,
                                 adjacency,
                                 region=region,
                                 impute_method=impute)
        print("CpG sites do not have the same trend in the same MR")
    else:
        out_df = data_handler4WGS(KO_df,
                                  WT_df,
                                  min_consecutive,
                                  adjacency,
                                  region=region,
                                  impute_method=impute)
        print("Use default setting, CpG sites do not have the same trend in the same MR")

    # Order
    out_df = out_df[['name', 'meth_points_KO', 'num_in_group_KO', 'orig_power_KOs',
                     'imputed_power_KOs', 'increased_power_KOs', 'meth_points_WT',
                     'num_in_group_WT', 'orig_power_WTs', 'imputed_power_WTs',
                     'increased_power_WTs', 'median_KO', 'median_WT', 'rratio']]

    if len(out_df) > 0:
        out_filenames = [out_folder+'/'+filename+'_imputedWith_'+str(impute)+'_overlapping_MRs.csv',
                         out_folder+'/'+filename+'_increasedSampleSize_overlapping_MRs.csv']
        out_df.drop(['increased_power_KOs', 'increased_power_WTs'], axis=1)\
              .to_csv(out_filenames[0], index=False)

        out_df.drop(['imputed_power_KOs', 'imputed_power_WTs'], axis=1)\
              .to_csv(out_filenames[1], index=False)


        if two_samples:
            out_KO, out_WT = clear_for_plot(out_df)
            single_out_filenames = [out_folder+'/'+KO_samplename+'_'+methylation+'_'+region+'_overlapping_MRs.csv',
                                    out_folder+'/'+WT_samplename+'_'+methylation+'_'+region+'_overlapping_MRs.csv']
            out_KO.to_csv(single_out_filenames[0], index=False)
            out_WT.to_csv(single_out_filenames[1], index=False)
        else:
            single_out_filenames = None

        return out_filenames, single_out_filenames
    else:
        return None, None
