import pandas as pd
import numpy as np
from .rawMeth_To_5mC_5hmC import get_numerical_chr as get_num_chr
import argparse

def make_BED_of_ref(df):
    """Creates BED formatted dataframe from input file and returns the dataframe

    Keyword arguments:
    df -- dataframe of reference"""

    df['gene_rna_name'] = df['gene_name'] + ':' + df['rna_name']
    df['dot'] = '.'
    df_BED = pd.DataFrame(df[['chr', 'start_pos', 'end_pos', 'gene_rna_name',
                              'dot', 'strand']])
    return df_BED

def find_uniq_genes(df, num_chr_name, human, remove_mir):
    """Filters and orders data and writes to file

    Keyword arguments:
    df -- pandas dataframe
    num_chr_name -- numerical chromosome name?
    human -- data from human? if False -> mouse or rat
    remove_mir -- remove genes with names starting with mir? (default True)
    """
    df_copy = df.copy()

    if remove_mir:
        #removes rows starting with case insensitive 'mir':
        df_copy = df_copy[df_copy.gene_name.str.lower().str.startswith('mir') == False]

    #removes rows where chr contains '_':
    df_copy = df_copy[df_copy.chr.str.contains('_') == False]

    #removes duplicates
    df_copy = df_copy.drop_duplicates(subset=['chr', 'strand', 'TSS', 'TES'])


    chrs = df_copy.chr.str[3:]

    X_mask = chrs == 'X'
    chrs.loc[X_mask] = get_num_chr(human, 'X')

    Y_mask = chrs == 'Y'
    chrs.loc[Y_mask] = get_num_chr(human, 'Y')

    M_mask = chrs == 'M'
    chrs.loc[M_mask] = get_num_chr(human, 'M')

    df_copy['order_by_chr'] = chrs

    if num_chr_name:
        df_copy.chr = pd.to_numeric(df_copy.order_by_chr)

        #sorts first by chr, then TSS, then rna_name
        df_copy = df_copy.sort_values(['chr', 'TSS', 'rna_name'], ascending=True)

    else:
        df_copy.order_by_chr = pd.to_numeric(df_copy.order_by_chr)
        df_copy = df_copy.sort_values(['order_by_chr', 'TSS', 'rna_name'], ascending=True)

    return df_copy

def main(reference, human, numerical_chr_name, remove_mir, folder_out):
    """Cleans reference file (refFlat)"""

    if human.lower() in ('yes', 'y'):
        human = 'human'
    elif human.lower() in ('mouse', 'm'):
        human = 'mouse'
    elif human.lower() in ('rat', 'r'):
        human = 'rat'
    else:
        print('***** ERROR in preprocess_reference.py: No info whether it is human or mouse. If human sample, enter [yes/y]. If mouse sample, enter [no/n]. Exiting. *****')
        exit(1)

    if numerical_chr_name.lower() in ('yes', 'y'):
        numerical_chr_name = True
    elif numerical_chr_name.lower() in ('no', 'n'):
        numerical_chr_name = False
    else:
        print('***** ERROR in preprocess_reference.py: numerical chromosome name must be whether [yes/y] or [no/n]. Exiting. *****')
        exit(1)

    if remove_mir.lower() in ('yes', 'y'):
        remove_mir = True
    elif remove_mir.lower() in ('no', 'n'):
        remove_mir = False
    else:
        print(remove_mir)
        print('***** ERROR: removeMir must be either [yes/y] or [no/n]. Exiting. *****')
        exit(1)
    outfilename = reference.split('/')[-1][:-4]


    df = pd.read_csv(reference, sep='\t', usecols=[0, 1, 2, 3, 4, 5],
                names=['gene_name', 'rna_name', 'chr', 'strand', 'TSS', 'TES'])

    new_ref = find_uniq_genes(df, numerical_chr_name, human, remove_mir)

    new_ref.rename(columns={'TSS': 'start_pos', 'TES': 'end_pos'}, inplace=True)
    ref_BED_df = make_BED_of_ref(new_ref)

    # Makes sure the columns are in this exact order
    new_ref = new_ref[['rna_name', 'chr', 'strand', 'start_pos', 'end_pos', 'gene_name']]
    ref_BED_df = ref_BED_df[['chr', 'start_pos', 'end_pos', 'gene_rna_name', 'dot', 'strand']]

    # Keeps header
    new_ref.to_csv(folder_out+'/'+outfilename+'_clean_sorted.txt', sep='\t', index=False)

    out_bed = folder_out+'/'+outfilename+'_clean_sorted.bed'
    # No header
    ref_BED_df.to_csv(out_bed, sep='\t', index=False, header=None)

    return out_bed
