import pandas as pd
import numpy as np
import os
import time
import argparse

def get_numerical_chr(human, string):
    """Returns the numbers of the respective chromosome names

    Keyword arguments:
    human -- is the data from human sample or mouse or rat ? Yes/mouse/rat
    string -- X, Y or M"""
    if human =='human':
        if string == 'X':
            return '23'
        elif string == 'Y':
            return '24'
        elif string == 'M':
            return '25'
    elif human =='mouse ':   # mouse
        if string == 'X':
            return '20'
        elif string == 'Y':
            return '21'
        elif string == 'M':
            return '22'
    elif human =='rat': #rat
        if string == 'X':
            return '21'
        elif string == 'Y':
            return '22'
        elif string == 'M':
            return '23'


def sort_bed_df(df, human, num_chr_name):
    """Sorting dataframe on chromosome name, either numerical or not,
    then on gene start, then on gene end

    Keyword arguments:
    df -- dataframe
    human -- is the data from human sample or mouse? Yes/mouse/rat
    num_chr_name -- numerical chromosome name? True/False
    """
    try:
        int(df.chr[0])
        chrs = df.chr
    except ValueError:
        chrs = df.chr.str[3:]

        X_mask = chrs == 'X'
        chrs.loc[X_mask] = get_numerical_chr(human, 'X')

        Y_mask = chrs == 'Y'
        chrs.loc[Y_mask] = get_numerical_chr(human, 'Y')

        M_mask = chrs == 'M'
        chrs.loc[M_mask] = get_numerical_chr(human, 'M')

    df['order_by_chr'] = chrs

    if num_chr_name:
        df.chr = pd.to_numeric(df.order_by_chr, downcast='signed')

        #sorts first by chr, then gene start, then gene end
        df = df.sort_values(['chr', 'chrStart', 'chrEnd'], ascending=True)

    else:
        df.order_by_chr = pd.to_numeric(df.order_by_chr)
        df = df.sort_values(['order_by_chr', 'chrStart', 'chrEnd'], ascending=True)

    df = df.drop(['order_by_chr'], axis=1)
    return df

def make_bed_files_sort(df, human, num_chr_name, bp):
    """Creates five .bed-files: 5mC, 5hmC, MspI, HpaII and BGT.

    Keyword arguments:
    df -- pandas dataframe
    human -- Yes/mouse/rat
    num_chr_name -- numerical chromosome name? True/False
    bp -- extension from cg position
    """

    df = pd.DataFrame({'chr': df['chr'], 'chrStart': df['site'],
                       'chrEnd': df['site'], 'MspI': df['MspI'],
                       'HpaII': df['HpaII'], 'BGT': df['BGT']})
    #Sorting
    df_sorted = sort_bed_df(df, human, num_chr_name)

    df_sorted['5mC'] = df_sorted['BGT']/df['HpaII']
    df_sorted['5hmC'] = df_sorted['MspI']/df['BGT']

    MspI_df = df_sorted[['chr', 'chrStart', 'chrEnd', 'MspI']].copy()
    HpaII_df = df_sorted[['chr', 'chrStart', 'chrEnd', 'HpaII']].copy()
    BGT_df = df_sorted[['chr', 'chrStart', 'chrEnd', 'BGT']].copy()
    five_mC_df = df_sorted[['chr', 'chrStart', 'chrEnd', '5mC']].copy()
    five_hmC_df = df_sorted[['chr', 'chrStart', 'chrEnd', '5hmC']].copy()
    five_mC_df['chrEnd']+=bp
    five_mC_df['chrStart']-=bp
    five_hmC_df['chrEnd']+=bp
    five_hmC_df['chrStart']-=bp

    #To make sure the df-columns are written to files in correct order
    MspI_df = MspI_df[['chr', 'chrStart', 'chrEnd', 'MspI']]
    HpaII_df = HpaII_df[['chr', 'chrStart', 'chrEnd', 'HpaII']]
    BGT_df = BGT_df[['chr', 'chrStart', 'chrEnd', 'BGT']]
    five_mC_df = five_mC_df[['chr', 'chrStart', 'chrEnd', '5mC']]
    five_hmC_df = five_hmC_df[['chr', 'chrStart', 'chrEnd', '5hmC']]

    return MspI_df, HpaII_df, BGT_df, five_mC_df, five_hmC_df

def make_bed_file_sort_methylated(df, human, num_chr_name):
    #Sorting
    df_sorted = sort_bed_df(df, human, num_chr_name)
    return df_sorted

def main(raw_file, human, numerical_chr_name, bp, methylated, gzip, type, folder_out):
    if human.lower() in ('yes', 'y'):
        human = 'human'
    elif human.lower() in ('mouse', 'm'):
        human = 'mouse'
    elif human.lower() in ('rat','r'):
        human='rat'
    else:
        print('***** ERROR in rawMeth_To_5mC_5hmC.py: No info whether it is human or mouse. If human sample, enter [yes/y]. If mouse sample, enter [no/n]. Exiting. *****')
        exit(1)

    if numerical_chr_name.lower() in ('yes', 'y'):
        numerical_chr_name = True
    elif numerical_chr_name.lower() in ('no', 'n'):
        numerical_chr_name = False
    else:
        print('***** ERROR in rawMeth_To_5mC_5hmC.py: numerical chromosome name must be whether [yes/y] or [no/n]. Exiting. *****')
        exit(1)

    if gzip:
        if type is not None:
            samplename = type+'.'+raw_file.split('/')[-1][:-7]
        else:
            samplename = raw_file.split('/')[-1][:-7]
    else:
        if type is not None:
            samplename = type+'.'+raw_file.split('/')[-1][:-4]
        else:
            samplename = raw_file.split('/')[-1][:-4]

    if methylated:
        if gzip:
            methylation_df = pd.read_csv(raw_file, sep='\t', usecols=[0,1,2,3],
                                     names=('chr', 'chrStart', 'chrEnd', 'level'),
                                     compression='gzip')
        else:
            methylation_df = pd.read_csv(raw_file, sep='\t', usecols=[0,1,2,3],
                                     names=('chr', 'chrStart', 'chrEnd', 'level'))

        df = make_bed_file_sort_methylated(methylation_df, human, numerical_chr_name)
        out_filenames = [folder_out+'/'+samplename+'_5mC.bed']
        #donot export header jbw
        df.to_csv(out_filenames[0], sep='\t', index=False, header=None)
        return out_filenames

    else:
        if gzip:
            methylation_df = pd.read_csv(raw_file, sep='\t', skiprows=[0],
                                     names=('chr', 'site', 'HpaII', 'BGT', 'MspI'),
                                     compression='gzip')
        else:
            methylation_df = pd.read_csv(raw_file, sep='\t', skiprows=[0],
                                     names=('chr', 'site', 'HpaII', 'BGT', 'MspI'))

        all_dfs = make_bed_files_sort(methylation_df, human, numerical_chr_name, bp)
        MspI_df, HpaII_df, BGT_df, five_mC_df, five_hmC_df = all_dfs

        out_filenames = [folder_out+'/'+samplename+'_MspI.bed',
                        folder_out+'/'+samplename+'_HpaII.bed',
                        folder_out+'/'+samplename+'_BGT.bed',
                        folder_out+'/'+samplename+'_5mC.bed',
                        folder_out+'/'+samplename+'_5hmC.bed']

        # Header is in output
        MspI_df.to_csv(out_filenames[0], sep='\t', index=False)
        HpaII_df.to_csv(out_filenames[1], sep='\t', index=False)
        BGT_df.to_csv(out_filenames[2], sep='\t', index=False)
        #donot export header jbw
        five_mC_df.to_csv(out_filenames[3], sep='\t', index=False,header=None)
        five_hmC_df.to_csv(out_filenames[4], sep='\t', index=False,header=None)

        return out_filenames
